"""Main application logic of khard including command line handling"""

from argparse import Namespace
import datetime
from email import message_from_string
from email.policy import SMTP as SMTP_POLICY
from email.headerregistry import Address, AddressHeader, Group
import locale
import logging
import operator
import os
import sys
import textwrap
from typing import cast, Callable, Iterable, Optional, Union

from . import helpers
from .address_book import AddressBookCollection, VdirAddressBook
from .exceptions import AddressBookNameError, AddressBookParseError, Cancelled
from .contacts import Contact
from . import cli
from .config import Config
from .formatter import Formatter
from .helpers import interactive
from .helpers.interactive import confirm
from .query import AndQuery, AnyQuery, OrQuery, Query, TermQuery
from .version import version as khard_version
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


logger = logging.getLogger(__name__)
config: Config
# the types that sys.exit() can digest
ExitStatus = Union[str, int, None]


def version_check(contact: Contact, description: str) -> bool:
    if contact.version not in config.supported_vcard_versions:
        print("Warning:\nThe {} is based on vcard version {} but khard only "
              "supports the modification of vcards with version 3.0 and 4.0.\n"
              "If you proceed, the contact will be converted to vcard version "
              "{} but beware: This could corrupt the contact file or cause "
              "data loss.".format(description, contact.version,
                                  config.preferred_vcard_version))
        if not confirm("Do you want to proceed anyway?"):
            print("Canceled")
            return False
    return True


def create_new_contact(address_book: VdirAddressBook) -> None:
    editor = interactive.Editor(config.editor, config.merge_editor)
    # create temp file
    template = "# create new contact\n# Address book: {}\n# Vcard version: " \
        "{}\n# if you want to cancel, exit without saving\n\n{}".format(
            address_book, config.preferred_vcard_version,
            helpers.get_new_contact_template(config.private_objects))
    new_contact = editor.edit_templates(lambda t: Contact.from_yaml(
        address_book, t, config.private_objects,
        config.preferred_vcard_version, config.localize_dates), template)

    # create contact object from temp file
    if new_contact is None:
        print("Canceled")
    else:
        new_contact.write_to_file()
        print("Creation successful\n\n{}".format(new_contact.pretty()))


def modify_existing_contact(old_contact: Contact) -> None:
    editor = interactive.Editor(config.editor, config.merge_editor)
    # create temp file and open it with the specified text editor
    text = ("# Edit contact: {}\n# Address book: {}\n# Vcard version: {}\n"
            "# if you want to cancel, exit without saving\n\n{}".format(
                old_contact, old_contact.address_book, old_contact.version,
                old_contact.to_yaml()))
    new_contact = editor.edit_templates(
        lambda t: Contact.clone_with_yaml_update(
            old_contact, t, config.localize_dates), text)

    # check if the user changed anything
    if new_contact is None or old_contact == new_contact:
        print("Nothing changed\n\n{}".format(old_contact.pretty()))
    else:
        new_contact.write_to_file(overwrite=True)
        print("Modification successful\n\n{}".format(new_contact.pretty()))


def merge_existing_contacts(source_contact: Contact,
                            target_contact: Contact,
                            delete_source_contact: bool) -> None:
    # show warning, if target vCard version is not 3.0 or 4.0
    if not version_check(target_contact, "target contact in which to merge"):
        return
    # create temp files for each vCard
    editor = interactive.Editor(config.editor, config.merge_editor)
    src_text = ("# merge from {}\n# Address book: {}\n# Vcard version: {}\n"
                "# if you want to cancel, exit without saving\n\n{}".format(
                    source_contact, source_contact.address_book,
                    source_contact.version, source_contact.to_yaml()))
    target_text = ("# merge into {}\n# Address book: {}\n# Vcard version: {}\n"
                   "# if you want to cancel, exit without saving\n\n{}".format(
                       target_contact, target_contact.address_book,
                       target_contact.version, target_contact.to_yaml()))
    merged_contact = editor.edit_templates(
        lambda t: Contact.clone_with_yaml_update(
            target_contact, t, config.localize_dates), src_text, target_text)
    # compare them
    if merged_contact is None or target_contact == merged_contact:
        raise Cancelled(
            f"Target contact unmodified\n\n{target_contact.pretty()}", 0)

    print("Merge contact {} from address book {} into contact {} from address "
          "book {}\n\n".format(source_contact, source_contact.address_book,
                               merged_contact, merged_contact.address_book))
    if delete_source_contact:
        print("To be removed")
    else:
        print("Keep unchanged")
    print("\n\n{}\n\nMerged\n\n{}\n".format(source_contact.pretty(),
                                            merged_contact.pretty()))
    if not confirm("Are you sure?"):
        print("Canceled")
        return

    # save merged_contact to disk and delete source contact
    merged_contact.write_to_file(overwrite=True)
    if delete_source_contact:
        source_contact.delete_vcard_file()
    print("Merge successful\n\n{}".format(merged_contact.pretty()))


def copy_contact(contact: Contact, target_address_book: VdirAddressBook,
                 delete_source_contact: bool) -> None:
    source_contact_filename = ""
    if delete_source_contact:
        # if source file should be moved, get its file location to delete after
        # successful movement
        source_contact_filename = contact.filename
    if not delete_source_contact or not contact.uid:
        # if copy contact or contact has no uid yet
        # create a new uid
        contact.uid = helpers.get_random_uid()
    # set destination file name
    contact.filename = os.path.join(target_address_book.path,
                                    "{}.vcf".format(contact.uid))
    # save
    contact.write_to_file()
    # delete old file
    if os.path.isfile(source_contact_filename):
        os.remove(source_contact_filename)
    print("{} contact {} from address book {} to {}".format(
        "Moved" if delete_source_contact else "Copied", contact,
        contact.address_book, target_address_book))


def list_address_books(address_books: Union[AddressBookCollection,
                                            list[VdirAddressBook]]) -> None:
    table = [["Index", "Address book"]]
    for index, address_book in enumerate(address_books, 1):
        table.append([cast(str, index), address_book.name])
    print(helpers.pretty_print(table))


def list_contacts(vcard_list: list[Contact], fields: Iterable[str] = (),
                  parsable: bool = False) -> None:
    selected_address_books: list[VdirAddressBook] = []
    selected_kinds = set()
    for contact in vcard_list:
        if contact.address_book not in selected_address_books:
            selected_address_books.append(contact.address_book)
        if contact.kind not in selected_kinds:
            selected_kinds.add(contact.kind)
    table = []
    # default table header
    table_header = ["index", "name", "phone", "email"]
    plural = ""
    if config.show_kinds or len(selected_kinds) > 1 or Contact._default_kind not in selected_kinds:
        table_header.append("kind")
    if len(selected_address_books) > 1:
        plural = "s"
        table_header.append("address_book")
    if not parsable:
        print("Address book{}: {}".format(plural, ', '.join(
                str(book) for book in selected_address_books)))
    if config.show_uids:
        table_header.append("uid")

    if parsable:
        # Legacy default header fields for parsable.
        table_header = ["uid", "name", "address_book"]

    if fields:
        table_header = [x.lower().replace(' ', '_') for x in fields]

    abook_collection = AddressBookCollection('short uids collection',
                                             selected_address_books)

    if not parsable:
        table.append([x.title().replace('_', ' ') for x in table_header])
    # table body
    formatter = Formatter(config.display, config.preferred_email_address_type,
                          config.preferred_phone_number_type,
                          config.show_nicknames, parsable)
    for index, vcard in enumerate(vcard_list):
        row = []
        for field in table_header:
            if field == 'index':
                row.append(str(index + 1))
            elif field in ['name', 'phone', 'email', 'kind']:
                row.append(formatter.get_special_field(vcard, field))
            elif field == 'uid':
                if parsable:
                    row.append(vcard.uid or "")
                elif uid := abook_collection.get_short_uid(vcard.uid or ""):
                    row.append(uid)
                else:
                    row.append("")
            else:
                row.append(formatter.get_nested_field(vcard, field))
        if parsable:
            print("\t".join(str(v) for v in row))
        else:
            table.append(row)
    if not parsable:
        print(helpers.pretty_print(table))


def list_with_headers(the_list: list[str], *headers: str) -> None:
    table = [list(headers)]
    for row in the_list:
        table.append(row.split("\t"))
    print(helpers.pretty_print(table))


def choose_address_book_from_list(header: str, abooks: Union[
                                  AddressBookCollection, list[VdirAddressBook]]
                                  ) -> Optional[VdirAddressBook]:
    """Let the user select one of the given address books

    :param header: some text to print in front of the list
    :param abooks: the address books from which to select
    :returns: the selected address book
    :raises Canceled: when the user canceled the selection
    """
    if not abooks:
        return None
    if len(abooks) == 1:
        return abooks[0]
    print(header)
    list_address_books(abooks)
    return interactive.select(abooks)


def choose_vcard_from_list(header: str, vcards: list[Contact],
                           include_none: bool = False
                           ) -> Optional[Contact]:
    """Let the user select a contact from a list

    :param header: some text to print in front of the list
    :param vcards: the contacts from which to select
    :returns: the selected contact
    :raises Canceled: when the user canceled the selection
    """
    if not vcards:
        return None
    if len(vcards) == 1 and not include_none:
        return vcards[0]
    print(header)
    list_contacts(vcards)
    return interactive.select(vcards, True)


def get_contact_list(address_books: Union[VdirAddressBook,
                                          AddressBookCollection],
                     query: Query) -> list[Contact]:
    """Find contacts in the given address book grouped, sorted and reversed
    according to the loaded configuration.

    :param address_books: the address book to search
    :param query: the query to use when searching
    :returns: list of found Contact objects
    """
    contacts = address_books.search(query)
    return sort_contacts(contacts, config.reverse, config.group_by_addressbook,
                         config.sort)


def sort_contacts(contacts: Iterable[Contact], reverse: bool = False,
                  group: bool = False, sort: str = "first_name") -> list[
                      Contact]:
    """Sort a list of contacts

    :param contacts: the contact list to sort
    :param reverse: reverse the order of the returned contacts
    :param group: group results by address book
    :param sort: the field to use for sorting, one of "first_name",
        "last_name", "formatted_name"
    :returns: sorted contact list
    """
    keys: list[Callable] = []
    if group:
        keys.append(operator.attrgetter("address_book.name"))
    if sort == "first_name":
        keys.append(operator.methodcaller("get_first_name_last_name"))
    elif sort == "last_name":
        keys.append(operator.methodcaller("get_last_name_first_name"))
    elif sort == "formatted_name":
        keys.append(operator.attrgetter("formatted_name"))
    else:
        raise ValueError('sort must be "first_name", "last_name" or '
                         '"formatted_name" not {}.'.format(sort))
    return sorted(contacts, reverse=reverse,
                  key=lambda x: [locale.strxfrm(key(x).lower())
                                 for key in keys])


def prepare_search_queries(args: Namespace) -> dict[str, Query]:
    """Prepare the search query string from the given command line args.

    Each address book can get a search query string to filter vCards before
    loading them.  Depending on the question if the address book is used for
    source or target searches different queries have to be combined.

    :param args: the parsed command line
    :returns: a dict mapping abook names to their loading queries
    """
    # get all possible search queries for address book parsing
    source_queries: list[Query] = []
    target_queries: list[Query] = []
    if "source_search_terms" in args:
        source_queries.append(args.source_search_terms)
    if "search_terms" in args:
        source_queries.append(args.search_terms)
    if "target_contact" in args:
        target_queries.append(args.target_contact)
    source_query = AndQuery.reduce(source_queries)
    target_query = AndQuery.reduce(target_queries)
    logger.debug('Created source query: %s', source_query)
    logger.debug('Created target query: %s', target_query)
    # Get all possible search queries for address book parsing, always
    # depending on the fact if the address book is used to find source or
    # target contacts or both.
    queries: dict[str, list[Query]] = {
        abook.name: [] for abook in config.abooks}
    for name in queries:
        if "addressbook" in args and name in args.addressbook:
            queries[name].append(source_query)
        if "target_addressbook" in args and name in args.target_addressbook:
            queries[name].append(target_query)
    queries2: dict[str, Query] = {
        n: OrQuery.reduce(q) for n, q in queries.items()}
    logger.debug('Created query: %s', queries)
    return queries2


def generate_contact_list(args: Namespace) -> list[Contact]:
    """Find the contact list with which we will work later on

    :param args: the command line arguments
    :returns: the contacts for further processing
    """
    if "source_search_terms" in args:
        # exception for merge command
        args.search_terms = args.source_search_terms or AnyQuery()
    if "search_terms" not in args:
        # It is simpler to handle subcommand that do not have and need search
        # terms here than conditionally calling generate_contact_list().
        return []
    return get_contact_list(args.addressbook, args.search_terms)


def new_subcommand(abooks: AddressBookCollection, data: str, open_editor: bool
                   ) -> ExitStatus:
    """Create a new contact.

    :param abooks: a list of address books that were selected on the command
        line
    :param data: the data for the new contact as a yaml formatted string
    :param open_editor: whether to open the new contact in the editor after
        creation
    :raises Canceled: when the user canceled a selection
    """
    # ask for address book, in which to create the new contact
    abook = choose_address_book_from_list(
        "Select address book for new contact", abooks)
    if abook is None:
        return "address book list is empty"
    # if there is some data in stdin/the input file
    if data:
        # create new contact from stdin/the input file
        try:
            new_contact = Contact.from_yaml(
                abook, data, config.private_objects,
                config.preferred_vcard_version, config.localize_dates)
        except ValueError as err:
            return str(err)
        else:
            new_contact.write_to_file()
        if open_editor:
            modify_existing_contact(new_contact)
        else:
            print("Creation successful\n\n{}".format(new_contact.pretty()))
    else:
        create_new_contact(abook)


def add_email_to_contact(name: str, email_address: str,
        abooks: AddressBookCollection, skip_already_added: bool) -> None:
    """Add a new email address to the given contact,
    creating the contact if necessary.

    :param name: name of the contact
    :param email_address: email address of the contact
    :param abooks: the address books that were selected on the command line
    :param skip_already_added: skip if email_address is part of one or more contacts
    :raises Canceled: when the user canceled a selection
    """

    # email address
    # search in contacts
    matching_contact_list = get_contact_list(abooks, TermQuery(email_address))
    if matching_contact_list:
        matching_contact_list_to_string = ', '.join(
                str(i) for i in matching_contact_list)
        if skip_already_added:
            print("Skipping email address {}: Is already part of {}"
                  .format(email_address, matching_contact_list_to_string))
            return
        if not confirm("Email address: {}, Found in contacts: {}. Select anyway?"
                       .format(email_address, matching_contact_list_to_string)):
            return
    else:
        if name:
            name_and_email = '"{}" <{}>'.format(name, email_address)
        else:
            name_and_email = email_address
        if not confirm("New address: {}. Select?".format(name_and_email)):
            return

    # name
    if not name:
        # ask for name
        name = input("Contact's name: ")
    else:
        # remove chars: " '
        name = name.replace('"', '').replace('\'', '')
    # backup name for the "create new contact" function part below
    original_name = name

    # select contact
    previous_name = name
    previous_selected_vcard = None
    manual_search = False
    while True:
        query: Query
        # search for an existing contact
        name_parts = name.replace(',', '').split()
        if len(name_parts) == 0:
            query = AnyQuery()
        elif len(name_parts) == 1:
            query = TermQuery(name)
        else:
            term_query_list = [TermQuery(part) for part in name_parts]
            query = AndQuery(
                    term_query_list[0], term_query_list[1], *term_query_list[2:])
        found_vcard_list = get_contact_list(abooks, query)

        # select contact from list
        if manual_search:
            selected_vcard = choose_vcard_from_list(
                    "Select contact for the search term: {}".format(name),
                    found_vcard_list, include_none=True)
            if found_vcard_list and not selected_vcard:
                # contact selection cancelled
                # restore previous data
                name = previous_name
                selected_vcard = previous_selected_vcard
            manual_search = False
        else:
            selected_vcard = choose_vcard_from_list(
                    "Select contact for the found e-mail address",
                    found_vcard_list)

        break_outer = False
        while True:
            if selected_vcard is None:
                if found_vcard_list:
                    message = "Contact selection cancelled"
                else:
                    message = "Nothing found for '{}'".format(name)
                answer = interactive.ask(message, ["create", "search", "quit"])
            else:
                answer = interactive.ask(
                    "Contact selected: {}".format(selected_vcard),
                    ["yes", "create", "details", "search", "quit"],
                    """You can enter one of these choices:

                      yes      proceed with selected contact
                      create   create a new contact
                      details  show details of selected contact
                      search   search for a different contact
                      quit     abort
                    """)

            if selected_vcard:
                if answer == 'yes':
                    break_outer = True
                    break
                if answer == 'details':
                    print("\n{}".format(selected_vcard.pretty()))
                    continue
            if answer == 'create':
                selected_vcard = None
                break_outer = True
                break
            if answer == 'search':
                # save data
                previous_name = name
                previous_selected_vcard = selected_vcard
                # enter search string
                if original_name:
                    name = input("Search for contact [ENTER='{}' or -='']: "
                                 .format(original_name)) or original_name
                    if name == "-":
                        name = ""
                else:
                    name = input("Search for contact: ")
                manual_search = True
                break
            if answer == 'quit':
                print("Cancelled")
                return

        if break_outer:
            # restore name
            name = original_name
            break

    # create new contact
    if selected_vcard is None:
        # first and last name variables
        name_parts = name.split()
        # detect format: last_name, first_name in name variable
        if name.count(",") == 1 \
                and len(name_parts) > 1 \
                and name_parts[0].endswith(","):
            # remove "," from presumed last name
            name_parts[0] = name_parts[0].replace(',', '')
            # put last_name to the list end
            name_parts.append(name_parts.pop(0))
        # fill variables
        first = name_parts[0] if len(name_parts) > 0 else name
        last = name_parts[-1] if len(name_parts) > 1 else ""

        # ask for address book, in which to create the new contact
        if not config.abooks:
            raise Cancelled("address book list is empty")
        else:
            selected_address_book = choose_address_book_from_list(
                    "Select address book for new contact", config.abooks)
            if selected_address_book is None:
                print("No address book selected")
                return

        # ask for name and organisation of new contact
        while True:
            if first:
                first_name = input("First name [ENTER='{}' or -='']: "
                                   .format(first)) or first
                if first_name == "-":
                    first_name = ""
            else:
                first_name = input("First name: ")

            if last:
                last_name = input("Last name [ENTER='{}' or -='']: "
                                  .format(last)) or last
                if last_name == "-":
                    last_name = ""
            else:
                last_name = input("Last name: ")

            if name and not first_name and not last_name:
                # first and last names are empty, maybe it's an organisation
                organisation = input("Organisation [ENTER='{}' or -='']: "
                                     .format(name)) or name
                if organisation == "-":
                    organisation = ""
            else:
                organisation = input("Organisation: ")

            if not first_name and not last_name and not organisation:
                print("Error: All fields are empty.")
            else:
                print("")
                break

        # create contact
        #
        # build template
        template_data = list()
        if first_name:
            template_data.append("First name   : {}".format(first_name))
        if last_name:
            template_data.append("Last name    : {}".format(last_name))
        if organisation:
            template_data.append("Organisation : {}".format(organisation))
        # confirm contact creation
        print("Verify input data\n{}"
              .format(textwrap.indent('\n'.join(template_data), 2*' ')))
        if not confirm("Create contact?", False):
            print("Cancelled")
            return
        selected_vcard = Contact.from_yaml(
                selected_address_book, '\n'.join(template_data),
                config.private_objects, config.preferred_vcard_version,
                config.localize_dates)
        if not selected_vcard:
            print("Could not create contact")
            return
        print("Contact created successfully")

    # check if the contact already contains the email address
    for _, email_list in sorted(selected_vcard.emails.items(),
                                key=lambda k: k[0].lower()):
        for email in email_list:
            if email == email_address:
                print("The contact {} already contains the email address {}"
                      .format(selected_vcard, email_address))
                return

    # ask for the email label
    print("\nAdding email address {} to contact {}\n"
          "Enter email label\n"
          "    vcard 3.0: At least one of home, internet, pref, work, x400\n"
          "    vcard 4.0: At least one of home, internet, pref, work\n"
          "    Or a custom label (only letters)".format(email_address,
                                                        selected_vcard))
    while True:
        label = input("email label [internet]: ") or "internet"
        try:
            selected_vcard.add_email(label, email_address)
        except ValueError as err:
            print(f"Error: {err}")
        else:
            break
    # save to disk
    selected_vcard.write_to_file(overwrite=True)
    print("Done.\n\n{}".format(selected_vcard.pretty()))


def find_email_addresses(text: str, fields: list[str]) -> list[Address]:
    """Search the text for email addresses in the given fields.

    :param text: the text to search for email addresses
    :param fields: the fields to look in for email addresses.
        The `all` field searches all headers.
    """
    message = message_from_string(text, policy=SMTP_POLICY)

    def extract_addresses(header) -> list[Address]:
        if header and isinstance(header, (AddressHeader, Group)):
            return list(header.addresses)
        return []

    email_addresses = []

    _all = any([f == "all" for f in fields])
    if _all:
        for _, value in message.items():
            email_addresses.extend(extract_addresses(value))
    else:
        for field in fields:
            email_addresses.extend(extract_addresses(message[field]))

    return email_addresses


def add_email_subcommand(
        text: str,
        abooks: AddressBookCollection,
        fields: list[str],
        skip_already_added: bool) -> ExitStatus:
    """Add a new email address to contacts, creating new contacts if necessary.

    :param text: the input text to search for the new email
    :param abooks: the address books that were selected on the command line
    :param field: the header field to extract contacts from
    :param skip_already_added: skip already known email addresses
    :raises Canceled: when the user canceled a selection
    """
    email_addresses = find_email_addresses(text, fields)
    if not email_addresses:
        return f"No email addresses found in fields {fields}"

    print("Khard: Add email addresses to contacts")

    for email_address in email_addresses:
        name = email_address.display_name
        address = email_address.addr_spec

        add_email_to_contact(name, address, abooks, skip_already_added)

        print()

    print("No more email addresses")


def birthdays_subcommand(vcard_list: list[Contact], parsable: bool
                         ) -> ExitStatus:
    """Print birthday contact table.

    :param vcard_list: the vCards to search for matching entries which should
        be printed
    :param parsable: machine readable output: columns divided by tabulator (\t)
    """
    # filter out contacts without a birthday date
    vcard_list = [vcard for vcard in vcard_list if vcard.birthday is not None]
    # sort by date (month and day)
    # The sort function should work for strings and datetime objects.  All
    # strings will be sorted before any datetime objects.
    vcard_list.sort(key=lambda x: (x.birthday.month, x.birthday.day)
                    if isinstance(x.birthday, datetime.datetime)
                    else (0, 0, x.birthday))
    # add to string list
    birthday_list: list[str] = []
    formatter = Formatter(config.display, config.preferred_email_address_type,
                          config.preferred_phone_number_type,
                          config.show_nicknames, parsable)
    for vcard in vcard_list:
        name = formatter.get_special_field(vcard, "name")
        if parsable:
            # We did filter out None above but the type checker does not know
            # this.
            bday = cast(Union[str, datetime.datetime], vcard.birthday)
            if isinstance(bday, str):
                date = bday
            else:
                date = bday.strftime("%Y.%m.%d")
            birthday_list.append("{}\t{}".format(date, name))
        else:
            date = vcard.get_formatted_birthday()
            birthday_list.append("{}\t{}".format(name, date))
    if birthday_list:
        if parsable:
            print('\n'.join(birthday_list))
        else:
            list_with_headers(birthday_list, "Name", "Birthday")
    else:
        if not parsable:
            print("Found no birthdays")
        return 1


def phone_subcommand(search_terms: Query, vcard_list: list[Contact],
        parsable: bool) -> ExitStatus:
    """Print a phone application friendly contact table.

    :param search_terms: used as search term to filter the contacts before
        printing
    :param vcard_list: the vCards to search for matching entries which should
        be printed
    :param parsable: machine readable output: columns divided by tabulator (\t)
    """
    formatter = Formatter(config.display, config.preferred_email_address_type,
                          config.preferred_phone_number_type,
                          config.show_nicknames, parsable)
    numbers: list[str] = []
    for vcard in vcard_list:
        field_line_list = []
        for type, number_list in sorted(vcard.phone_numbers.items(),
                                        key=lambda k: k[0].lower()):
            for number in sorted(number_list):
                name = formatter.get_special_field(vcard, "name")
                if parsable:
                    # parsable option: start with phone number
                    fields = number, name, type
                else:
                    # else: start with name
                    fields = name, type, number
                field_line_list.append("\t".join(fields))
        numbers += _filter_email_post_or_phone_number_results(
                search_terms, field_line_list)
    if numbers:
        if parsable:
            print('\n'.join(numbers))
        else:
            list_with_headers(numbers, "Name", "Type", "Phone")
    else:
        if not parsable:
            print("Found no phone numbers")
        return 1


def post_address_subcommand(search_terms: Query,
        vcard_list: list[Contact], parsable: bool
                            ) -> ExitStatus:
    """Print a contact table with all postal / mailing addresses

    :param search_terms: used as search term to filter the contacts before
        printing
    :param vcard_list: the vCards to search for matching entries which should
        be printed
    :param parsable: machine readable output: columns divided by tabulator (\t)
    """
    formatter = Formatter(config.display, config.preferred_email_address_type,
                          config.preferred_phone_number_type,
                          config.show_nicknames, parsable)
    addresses: list[str] = []
    for vcard in vcard_list:
        name = formatter.get_special_field(vcard, "name")
        # create post address line list
        field_line_list = []
        if parsable:
            for type, post_addresses in sorted(vcard.post_addresses.items(),
                                               key=lambda k: k[0].lower()):
                for post_address in post_addresses:
                    field_line_list.append(
                            "\t".join([str(post_address), name, type]))
        else:
            for type, formatted_addresses in sorted(
                    vcard.get_formatted_post_addresses().items(),
                    key=lambda k: k[0].lower()):
                for address in sorted(formatted_addresses):
                    field_line_list.append(
                            "\t".join([name, type, address]))
        addresses += _filter_email_post_or_phone_number_results(
                search_terms, field_line_list)
    if addresses:
        if parsable:
            print('\n'.join(addresses))
        else:
            list_with_headers(addresses, "Name", "Type", "Post address")
    else:
        if not parsable:
            print("Found no post addresses")
        return 1


def email_subcommand(search_terms: Query, vcard_list: list[Contact],
                     parsable: bool, remove_first_line: bool) -> ExitStatus:
    """Print a mail client friendly contacts table that is compatible with the
    default format used by mutt.
    Output format:

    .. code-block:: text

        single line of text
        email_address\tname\ttype
        email_address\tname\ttype
        [...]

    :param search_terms: used as search term to filter the contacts before
        printing
    :param vcard_list: the vCards to search for matching entries which should
        be printed
    :param parsable: machine readable output: columns divided by tabulator (\t)
    :param remove_first_line: remove first line (searching for '' ...)
    """
    formatter = Formatter(config.display, config.preferred_email_address_type,
                          config.preferred_phone_number_type,
                          config.show_nicknames, parsable)
    emails: list[str] = []
    for vcard in vcard_list:
        field_line_list = []
        for type, email_list in sorted(vcard.emails.items(),
                                       key=lambda k: k[0].lower()):
            for email in sorted(email_list):
                name = formatter.get_special_field(vcard, "name")
                if parsable:
                    # parsable option: start with email address
                    fields = email, name, type
                else:
                    # else: start with name
                    fields = name, type, email
                field_line_list.append("\t".join(fields))
        emails += _filter_email_post_or_phone_number_results(
                search_terms, field_line_list)
    if emails:
        if parsable:
            if not remove_first_line:
                # at least mutt requires that line
                print("searching for '{}' ...".format(search_terms))
            print('\n'.join(emails))
        else:
            list_with_headers(emails, "Name", "Type", "E-Mail")
    else:
        if not parsable:
            print("Found no email addresses")
        elif not remove_first_line:
            print("searching for '{}' ...".format(search_terms))
        return 1


def _filter_email_post_or_phone_number_results(search_terms: Query,
        field_line_list: list[str]) -> list[str]:
    """Filter the created output of phone_subcommand, post_address_subcommand
    and email_subcommand by the given search term again.
    If no match is found, return the complete input list

    :param search_terms: used as search term to filter the contacts before
        printing
    :param field_line_list: The line-by-line output of the commands listed above
    """
    matched_line_list = []
    for line in field_line_list:
        if search_terms and search_terms.match(line):
            matched_line_list.append(line)
    return matched_line_list if matched_line_list else field_line_list


def list_subcommand(vcard_list: list[Contact], parsable: bool,
                    fields: list[str]) -> ExitStatus:
    """Print a user friendly contacts table.

    :param vcard_list: the vCards to print
    :param parsable: machine readable output: columns divided by tabulator (\t)
    :param fields: list of strings for field evaluation
    """
    if not vcard_list:
        if not parsable:
            print("Found no contacts")
        return 1
    else:
        list_contacts(vcard_list, fields, parsable)


def modify_subcommand(selected_vcard: Contact,
                      input_from_stdin_or_file: str, open_editor: bool,
                      source: bool = False) -> ExitStatus:
    """Modify a contact in an external editor.

    :param selected_vcard: the contact to modify
    :param input_from_stdin_or_file: new data from stdin (or a file) that
        should be incorporated into the contact, this should be a yaml
        formatted string
    :param open_editor: whether to open the new contact in the editor after
        creation
    :param source: edit the source file or a yaml version?
    """
    if source:
        editor = interactive.Editor(config.editor, config.merge_editor)
        editor.edit_files(selected_vcard.filename)
        return None
    # show warning, if vCard version of selected contact is not 3.0 or 4.0
    if not version_check(selected_vcard, "selected contact"):
        return None
    # if there is some data in stdin
    if input_from_stdin_or_file:
        # create new contact from stdin
        try:
            new_contact = Contact.clone_with_yaml_update(
                selected_vcard, input_from_stdin_or_file,
                config.localize_dates)
        except ValueError as err:
            return str(err)
        if selected_vcard == new_contact:
            print("Nothing changed\n\n{}".format(new_contact.pretty()))
        else:
            print("Modification\n\n{}\n".format(new_contact.pretty()))
            if confirm("Do you want to proceed?"):
                new_contact.write_to_file(overwrite=True)
                if open_editor:
                    modify_existing_contact(new_contact)
                else:
                    print("Done")
            else:
                print("Canceled")
    else:
        modify_existing_contact(selected_vcard)


def remove_subcommand(selected_vcard: Contact, force: bool) -> None:
    """Remove a contact from the address book.

    :param selected_vcard: the contact to delete
    :param force: delete without confirmation
    """
    if not force and not confirm(
            "Deleting contact {} from address book {}. Are you sure?".format(
                selected_vcard, selected_vcard.address_book)):
        print("Canceled")
        return
    selected_vcard.delete_vcard_file()
    print("Contact {} deleted successfully".format(
        selected_vcard.formatted_name))


def merge_subcommand(vcards: list[Contact],
                     abooks: AddressBookCollection, search_terms: Query
                     ) -> ExitStatus:
    """Merge two contacts into one.

    :param vcards: the vCards from which to choose contacts for merging
    :param abooks: the address books to use to find the target contact
    :param search_terms: the search terms to find the target contact
    :raises Canceled: when the user canceled a selection
    """
    # Find possible target contacts.
    target_vcards = get_contact_list(abooks, search_terms)
    # get the source vCard, from which to merge
    source_vcard = choose_vcard_from_list("Select contact from which to merge",
                                          vcards)
    if source_vcard is None:
        return "Found no source contact for merging"
    else:
        print("Merge from {} from address book {}\n\n".format(
            source_vcard, source_vcard.address_book))
    # get the target vCard, into which to merge
    target_vcard = choose_vcard_from_list("Select contact into which to merge",
                                          target_vcards)
    if target_vcard is None:
        return "Found no target contact for merging"
    else:
        print("Merge into {} from address book {}\n\n".format(
            target_vcard, target_vcard.address_book))
    # merging
    if source_vcard == target_vcard:
        print("The selected contacts are already identical")
    else:
        merge_existing_contacts(source_vcard, target_vcard, True)


def copy_or_move_subcommand(action: str, vcards: list[Contact],
                            target_address_books: AddressBookCollection
                            ) -> ExitStatus:
    """Copy or move a contact to a different address book.

    :param action: the string "copy" or "move" to indicate what to do
    :param vcards: the contact list from which to select one for the action
    :param target_address_books: the target address books
    :raises Canceled: when the user canceled a selection
    """
    # get the source vCard, which to copy or move
    source_vcard = choose_vcard_from_list(
        "Select contact to {}".format(action.title()), vcards)
    if source_vcard is None:
        return "Found no contact"
    else:
        print("{} contact {} from address book {}".format(
            action.title(), source_vcard, source_vcard.address_book))

    # get target address book
    if len(target_address_books) == 1 \
            and target_address_books[0] == source_vcard.address_book:
        return f"The address book {target_address_books[0]} already contains the contact {source_vcard}"
    else:
        available_address_books = [abook for abook in target_address_books
                                   if abook != source_vcard.address_book]
        target_abook = choose_address_book_from_list(
            "Select target address book", available_address_books)
        if target_abook is None:
            return "address book list is empty"

    # check if a contact already exists in the target address book
    target_vcard = choose_vcard_from_list(
        "Select target contact to overwrite (or None to add a new entry)",
        get_contact_list(target_abook, TermQuery(source_vcard.formatted_name)),
        True)
    # If the target contact doesn't exist, move or copy the source contact into
    # the target address book without further questions.
    if target_vcard is None:
        copy_contact(source_vcard, target_abook, action == "move")
    elif source_vcard == target_vcard:
        # source and target contact are identical
        print("Target contact: {}".format(target_vcard))
        if action == "move":
            copy_contact(source_vcard, target_abook, True)
        else:
            print("The selected contacts are already identical")
    else:
        # source and target contacts are different
        # either overwrite the target one or merge into target contact
        print("The address book {} already contains the contact {}\n\n"
              "Source\n\n{}\n\nTarget\n\n{}\n\n".format(
              target_vcard.address_book, source_vcard, source_vcard.pretty(),
              target_vcard.pretty()))
        while True:
            answer = interactive.ask(
                "Possible actions", [action, "merge", "overwrite", "quit"],
                "quit")
            if answer == action:
                copy_contact(source_vcard, target_abook, action == "move")
                break
            if answer == "overwrite":
                copy_contact(source_vcard, target_abook, action == "move")
                target_vcard.delete_vcard_file()
                break
            if answer == "merge":
                merge_existing_contacts(source_vcard, target_vcard,
                                        action == "move")
                break
            if answer == "quit":
                print("Canceled")
                break


def main(argv: list[str] = sys.argv[1:]) -> ExitStatus:
    locale.setlocale(locale.LC_ALL, "")
    args, conf = cli.init(argv)

    # store the config instance in the module level variable
    global config
    config = conf

    # Check some of the simpler subcommands first.  These don't have any
    # options and can directly be run.
    if args.action == "addressbooks":
        print('\n'.join(str(book) for book in config.abooks))
        return None
    if args.action == "template":
        print("# Contact template for khard version {}\n#\n"
              "# Use this yaml formatted template to create a new contact:\n"
              "#   either with: khard new -a address_book -i template.yaml\n"
              "#   or with: cat template.yaml | khard new -a address_book\n"
              "\n{}".format(khard_version, helpers.get_new_contact_template(
                  config.private_objects)))
        return None

    search_queries = prepare_search_queries(args)

    # load address books
    try:
        if "addressbook" in args:
            args.addressbook = config.get_address_books(args.addressbook,
                                                        search_queries)
        if "target_addressbook" in args:
            args.target_addressbook = config.get_address_books(
                args.target_addressbook, search_queries)
    except AddressBookParseError as err:
        return f"{err}\nUse --debug for more information or --skip-unparsable to proceed"
    except AddressBookNameError as err:
        return str(err)

    vcard_list = generate_contact_list(args)

    if args.action == "filename":
        print('\n'.join(contact.filename for contact in vcard_list))
        return None

    # read from template file or stdin if available
    input_from_stdin_or_file = ""
    if "input_file" in args:
        if args.input_file != "-":
            # try to read from specified input file
            try:
                with open(args.input_file, "r") as infile:
                    input_from_stdin_or_file = infile.read()
            except OSError as err:
                return "Error: {}\n       File: {}".format(err.strerror, err.filename)
        elif not sys.stdin.isatty():
            # try to read from stdin
            try:
                input_from_stdin_or_file = sys.stdin.read()
            except OSError:
                return "Error: Can't read from stdin"
            # try to reopen console
            # otherwise further user interaction is not possible (for example
            # selecting a contact from the contact table)
            try:
                tty_name = 'CONIN$' if os.name == 'nt' else '/dev/tty'
                sys.stdin = open(tty_name, 'r')
            except OSError:
                pass

    # these listing commands do not require any user interaction
    if args.action == "birthdays":
        return birthdays_subcommand(vcard_list, args.parsable)
    elif args.action == "phone":
        return phone_subcommand(args.search_terms, vcard_list, args.parsable)
    elif args.action == "postaddress":
        return post_address_subcommand(args.search_terms, vcard_list, args.parsable)
    elif args.action == "email":
        return email_subcommand(args.search_terms, vcard_list, args.parsable,
                                args.remove_first_line)
    elif args.action == "list":
        return list_subcommand(vcard_list, args.parsable, args.fields)

    else:
        # these commands require user interaction
        try:
            if args.action == "new":
                return new_subcommand(args.addressbook, input_from_stdin_or_file,
                                      args.open_editor)
            elif args.action == "add-email":
                return add_email_subcommand(input_from_stdin_or_file,
                                     args.addressbook, args.headers,
                                     args.skip_already_added)
            elif args.action in ["show", "edit", "remove"]:
                selected_vcard = choose_vcard_from_list(
                    "Select contact for {} action".format(args.action.title()),
                    vcard_list)
                if selected_vcard is None:
                    return "Found no contact"
                if args.action == "show":
                    if args.format == "pretty":
                        output = selected_vcard.pretty()
                    elif args.format == "vcard":
                        output = open(selected_vcard.filename).read()
                    else:
                        output = "# Contact template for khard version {}\n" \
                                 "# Name: {}\n# Vcard version: {}\n\n{}".format(
                                     khard_version, selected_vcard,
                                     selected_vcard.version,
                                     selected_vcard.to_yaml())
                    args.output_file.write(output)
                elif args.action == "edit":
                    return modify_subcommand(selected_vcard, input_from_stdin_or_file,
                                      args.open_editor, args.format == 'vcard')
                elif args.action == "remove":
                    remove_subcommand(selected_vcard, args.force)
            elif args.action == "merge":
                return merge_subcommand(vcard_list, args.target_addressbook,
                                        args.target_contact)
            elif args.action in ["copy", "move"]:
                return copy_or_move_subcommand(args.action, vcard_list,
                                               args.target_addressbook)
        except Cancelled as ex:
            if ex.code == 0:
                print(ex)
            else:
                print(f"Error: {ex}")
            return ex.code



#!/usr/bin/env python

import datetime
import os
import subprocess

import config

# current date and time
datetime = datetime.datetime.now()
current_date = "%.2d.%.2d.%.4d" % (datetime.day, datetime.month, datetime.year)
current_time = "%.2d:%.2d:%.2d" % (datetime.hour, datetime.minute, datetime.second)

# if music was stopped, resume again
if os.path.exists(config.mpd_lockfile):
    os.remove(config.mpd_lockfile)
    subprocess.call(["mpc", "-h", config.mpd_host, "-p", str(config.mpd_port), "play"])

# try to get the caller name / id from the previously created temp file
try:
    with open(config.caller_id_filename, "r") as caller_id_file:
        caller_id = caller_id_file.read().strip()
except:
    caller_id = "anonymous"
if config.language == "de":
    message = "Anruf in Abwesenheit von %s am %s um %s\n" % (caller_id, current_date, current_time)
else:
    message = "Call in absence of %s in %s at %s\n" % (caller_id, current_date, current_time)
try:
    os.remove(config.caller_id_filename)
except:
    pass

# log into file
with open(config.call_log_file, "a") as log:
    log.write(message)



"""Helper functions for user interaction."""

import contextlib
from datetime import datetime
from enum import Enum
import os.path
import subprocess
from tempfile import NamedTemporaryFile
from typing import Callable, Generator, Optional, Sequence, TypeVar, Union

from ..exceptions import Cancelled
from ..contacts import Contact


T = TypeVar("T")


def confirm(message: str, accept_enter_key: bool = True) -> bool:
    """Ask the user for confirmation on the terminal.

    :param message: the question to print
    :param accept_enter_key: Accept ENTER as alternative for "n"
    :returns: the answer of the user
    """
    return "yes" == ask(message, ["yes", "no"],
                        "no" if accept_enter_key else None)


def ask(message: str, choices: list[str], default: Optional[str] = None,
        help: Optional[str] = None) -> str:
    """Ask the user to select one of the given choices

    :param message: a text to show to the user
    :param choices: the possible answers the user might give, if help is not
        None this list must not contain the string "?"
    :param default: the answer that should be selected on empty user input
        (None means empty input is not accepted)
    :param help: a help text to display to the user if they did not answer
        correctly
    :returns: the choice of the user
    """
    default = default.lower() if default is not None else None
    # ensure that the choices are lower case, in order but unique
    choices = list({c.lower(): None for c in choices})
    prompt = "/".join("[{}]".format(c) if c == default else c
                      for c in choices)
    if help is not None:
        prompt += " or ? for help"
    prompt += ": "
    if len(message) + len(prompt) < 79:
        prompt = message + " " + prompt
    else:
        print(message)
    while True:
        try:
            answer = input(prompt).lower()
            if answer == "" and default is not None:
                return default
            if answer == "?" and help is not None:
                print(help)
                continue
            if answer in choices:
                return answer
            prefixes_matches = [c for c in choices if c.startswith(answer)]
            if len(prefixes_matches) == 1:
                return prefixes_matches[0]
            if len(prefixes_matches) > 1:
                print("The given prefix is not specific enough.")
        except (EOFError, IndexError, ValueError):
            pass
        except KeyboardInterrupt:
            raise Cancelled
        if help is not None:
            print(help)


def select(items: Sequence[T], include_none: bool = False) -> Optional[T]:
    """Ask the user to select an item from a list.

    The list should be displayed to the user before calling this function and
    should be indexed starting with 1.

    :param items: the list from which to select
    :param include_none: whether to allow the selection of no item
    :returns: None or the selected item
    :raises Canceled: when the user canceled the selection process
    """
    prompt = "Enter Index ({}q to quit): ".format("0 for None, "
                                                  if include_none else "")
    while True:
        try:
            answer = input(prompt)
            answer = answer.lower()
            if answer == "q":
                raise Cancelled
            index = int(answer)
            if include_none and index == 0:
                return None
            if index > 0:
                return items[index - 1]
        except (EOFError, IndexError, ValueError):
            pass
        print("Please enter an index value between 1 and {} or q to quit."
              .format(len(items)))


class EditState(Enum):
    modified = 1
    unmodified = 2
    aborted = 3


class Editor:

    """Wrapper around subprocess.Popen to edit and merge files."""

    def __init__(self, editor: Union[str, list[str]],
                 merge_editor: Union[str, list[str]]) -> None:
        self.editor = [editor] if isinstance(editor, str) else editor
        self.merge_editor = [merge_editor] if isinstance(merge_editor, str) \
            else merge_editor

    @staticmethod
    @contextlib.contextmanager
    def write_temp_file(text: str = "") -> Generator[str, None, None]:
        """Create a new temporary file and write some initial text to it.

        :param text: the text to write to the temp file
        :returns: the file name of the newly created temp file
        """
        with NamedTemporaryFile(mode='w+t', suffix='.yml') as tmp:
            tmp.write(text)
            tmp.flush()
            yield tmp.name

    @staticmethod
    def _mtime(filename: str) -> datetime:
        return datetime.fromtimestamp(os.path.getmtime(filename))

    def edit_files(self, file1: str, file2: Optional[str] = None) -> EditState:
        """Edit the given files

        If only one file is given the timestamp of this file is checked, if two
        files are given the timestamp of the second file is checked for
        modification.

        :param file1: the first file (checked for modification if file2 not
            present)
        :param file2: the second file (checked for modification of present)
        :returns: the result of the modification
        """
        if file2 is None:
            command = self.editor + [file1]
        else:
            command = self.merge_editor + [file1, file2]
        timestamp = self._mtime(command[-1])
        child = subprocess.Popen(command)
        child.communicate()
        if child.returncode != 0:
            return EditState.aborted
        if timestamp == self._mtime(command[-1]):
            return EditState.unmodified
        return EditState.modified

    def edit_templates(self, yaml2card: Callable[[str], Contact],
                       template1: str, template2: Optional[str] = None
                       ) -> Optional[Contact]:
        """Edit YAML templates of contacts and parse them back

        :param yaml2card: a function to convert the modified YAML templates
            into a Contact
        :param template1: the first template
        :param template2: the second template (optional, for merges)
        :returns: the parsed Contact or None
        """
        templates = [t for t in (template1, template2) if t is not None]
        with contextlib.ExitStack() as stack:
            files = [stack.enter_context(self.write_temp_file(t))
                     for t in templates]
            # Try to edit the files until we detect a modification or the user
            # aborts
            while True:
                if self.edit_files(*files) == EditState.unmodified:
                    return None
                # read temp file contents after editing
                with open(files[-1], "r") as tmp:
                    modified_template = tmp.read()
                # No actual modification was done
                if modified_template == templates[-1]:
                    return None
                # try to create contact from user input
                try:
                    return yaml2card(modified_template)
                except ValueError as err:
                    print(f"\nError: {err}\n")
                    if not confirm("Do you want to open the editor again?"):
                        print("Canceled")
                        return None
        return None  # only for mypy



"""Names and aliases for the subcommands on the command line"""

from typing import Generator, Iterable, Optional


class Actions:

    """A class to manage the names and aliases of the command line
    subcommands."""

    action_map: dict[str, list[str]] = {
        "add-email":    [],
        "addressbooks": ["abooks"],
        "birthdays":    ["bdays"],
        "copy":         ["cp"],
        "email":        [],
        "filename":     ["file"],
        "list":         ["ls"],
        "merge":        [],
        "edit":         ["modify", "ed"],
        "move":         ["mv"],
        "new":          ["add"],
        "phone":        [],
        "postaddress":  ["post", "postaddr"],
        "remove":       ["delete", "del", "rm"],
        "show":         ["details"],
        "template":     [],
    }

    @classmethod
    def get_action(cls, alias: str) -> Optional[str]:
        """Find the name of the action for the supplied alias.  If no action is
        associated with the given alias, None is returned.

        :param alias: the alias to look up
        :returns: the name of the corresponding action or None

        """
        for action, alias_list in cls.action_map.items():
            if alias in alias_list:
                return action
        return None

    @classmethod
    def get_aliases(cls, action: str) -> list[str]:
        """Find all aliases for the given action.  If there is no such action,
        None is returned.

        :param action: the action name to look up
        :returns: the list of aliases corresponding to the action or None

        """
        return cls.action_map[action]

    @classmethod
    def get_actions(cls) -> Iterable[str]:
        """Find the names of all defined actions.

        :returns: all action names
        """
        return cls.action_map.keys()

    @classmethod
    def get_all(cls) -> Generator[str, None, None]:
        """Find the names of all defined actions and their aliases.

        :returns: the names of all actions and aliases
        """
        for key, value in cls.action_map.items():
            yield key
            yield from value



"""Custom exceptions for khard"""


class Cancelled(Exception):
    """An exception indicating that the user canceled some operation or
    some backend operation failed"""

    def __init__(self, message: str = "Canceled", code: int = 1) -> None:
        super().__init__(message)
        self.code = code


class AddressBookParseError(Exception):
    """Indicate an error while parsing data from an address book backend."""

    def __init__(self, filename: str, abook: str, reason: Exception) -> None:
        """Store the filename that caused the error."""
        super().__init__()
        self.filename = filename
        self.abook = abook
        self.reason = reason

    def __str__(self) -> str:
        return f"Error when parsing {self.filename} in address book {self.abook}: {self.reason}"


class AddressBookNameError(Exception):
    """Indicate an error with an address book name."""


class ConfigError(Exception):
    """Errors during config file parsing"""



#!/usr/bin/env python

# This script speaks the incoming caller ID for the SIP client Twinkle
# The following programs are needed: espeak, ffmpeg and sox, mpc is optional
# aptitude install ffmpeg espeak sox mpc
# Further information about Twinkle scripts can be found at
# http://mfnboer.home.xs4all.nl/twinkle/manual.html#profile_scripts

import os
import re
import subprocess
import sys

import config

def get_caller_id(from_hdr):
    caller_id = from_hdr[from_hdr.find(":")+1:from_hdr.find("@")]
    # remove all non digits from caller id
    caller_id = re.sub(r"\D", "", caller_id)
    # remove two digit country identification if present
    if not caller_id.startswith("0"):
        return caller_id[2:]
    return caller_id

def caller_from_addressbook(caller_id):
    try:
        callers = subprocess.check_output([config.khard_exe, "phone", "--parsable", caller_id]).strip()
    except subprocess.CalledProcessError:
        return caller_id
    if len(callers.split("\n")) == 1:
        return callers.split("\t")[1]
    else:
        # the contact contains multiple phone numbers and we have to obtain the right phone label
        regexp = re.compile(caller_id, re.IGNORECASE)
        for entry in callers.split("\n"):
            if regexp.search(re.sub("\D", "", entry.split("\t")[0])) is not None:
                return "%s (%s)" % (entry.split("\t")[1], entry.split("\t")[2])
        return callers.split("\n")[0].split("\t")[1]

def create_ringtone(caller_id):
    if os.path.exists(config.new_ringtone):
        os.remove(config.new_ringtone)
    if config.language == "de":
        subprocess.call(["espeak", "-v", "de", "-s", "300", "-w", config.tmp_mono_file, caller_id])
    else:
        subprocess.call(["espeak", "-v", "en-us", "-s", "300", "-w", config.tmp_mono_file, caller_id])
    subprocess.call(["ffmpeg", "-i", config.tmp_mono_file, "-ar", "48000", "-ac", "2", "-y", config.tmp_file_stereo],
            stdout=open('/dev/null', 'w'), stderr=open('/dev/null', 'w'))
    subprocess.call(["sox", config.constant_ringtone_segment, config.tmp_file_stereo, config.new_ringtone])


# main part of the script
if not os.path.exists(config.constant_ringtone_segment):
    print("The constant part of the ringtone file is missing. Create the sounds folder in your twinkle config and put a wav file in it")
    sys.exit(1)

# pause the music playback
# I use a MPD server for playing music so I pause it with the client MPC
# You can disable that in the config.py file
if config.stop_music:
    mpc_output = subprocess.check_output(["mpc", "-h", config.mpd_host, "-p", str(config.mpd_port), "status"])
    if "playing" in mpc_output:
        subprocess.call(["mpc", "-h", config.mpd_host, "-p", str(config.mpd_port), "pause"])
        music_tmp_file = open(config.mpd_lockfile, "w")
        music_tmp_file.close()

if "SIP_FROM" in os.environ:
    from_hdr = os.environ["SIP_FROM"]
    # parse the caller ID of the string
    caller_id = get_caller_id(from_hdr)
    # look into the address book
    if caller_id != "":
        caller_id = caller_from_addressbook(caller_id)
    else:
        caller_id = "anonymous"
    # create the ringtone
    if config.language == "de":
        create_ringtone("Anruf von " + caller_id)
    else:
        create_ringtone("Call from " + caller_id)
    # save the caller id for later use
    with open(config.caller_id_filename, "w") as caller_id_file:
        caller_id_file.write(caller_id)
    # if the file creation was successful and the file exists, tell twinkle to use it as the ringtone
    # else do nothing and play the standard ringtone
    if os.path.exists(config.new_ringtone):
        print("ringtone=" + config.new_ringtone)
sys.exit()



"""Classes and logic to handle vCards in khard.

This module explicitly supports the vCard specifications version 3.0 and 4.0
which can be found here:
- version 3.0: https://tools.ietf.org/html/rfc2426
- version 4.0: https://tools.ietf.org/html/rfc6350

For further details about the application and device values for the kind
field see:
- application: https://tools.ietf.org/html/rfc6473
- device: https://tools.ietf.org/html/rfc6869
"""

import contextlib
import copy
import datetime
import io
import locale
import logging
import os
import re
import tempfile
import time
from typing import Any, Callable, IO, Iterator, Literal, Optional, TypeVar, \
    Union, Sequence, overload

from ruamel import yaml
from ruamel.yaml import YAML
import vobject

from khard.exceptions import Cancelled

from . import address_book  # pylint: disable=unused-import # for type checking
from . import helpers
from .helpers.typing import (Date, PostAddress, StrList, convert_to_vcard,
    list_to_string, string_to_date, string_to_list)
from .query import AnyQuery, Query


logger = logging.getLogger(__name__)
T = TypeVar("T")
LabeledStrs = list[Union[str, dict[str, str]]]


@overload
def multi_property_key(item: str) -> tuple[Literal[0], str]: ...
@overload
def multi_property_key(item: dict[T, Any]) -> tuple[Literal[1], T]: ...
@overload
def multi_property_key(item: Union[str, dict[T, Any]]) -> tuple[Literal[0, 1], Union[T, str]]: ...
def multi_property_key(item: Union[str, dict[T, Any]]) -> tuple[Literal[0, 1], Union[T, str]]:
    """Key function to pass to sorted(), allowing sorting of dicts with lists
    and strings. Dicts will be sorted by their label, after other types.

    :param item: member of the list being sorted
    :type item: a dict with a single entry or any sortable type
    :returns: a pair, the first item is int(isinstance(item, dict).
        The second is either the key from the dict or the unchanged item if it
        is not a dict.
    """
    if isinstance(item, dict):
        return (1, next(iter(item)))
    return (0, item)


class VCardWrapper:
    """Wrapper class around a vobject.vCard object.

    This class can wrap a single vCard and presents its data in a manner
    suitable for khard.  Additionally some details of the vCard specifications
    in RFC 2426 (version 3.0) and RFC 6350 (version 4.0) that are not enforced
    by the vobject library are enforced here.
    """

    _default_version = "3.0"
    _supported_versions = ("3.0", "4.0")
    _default_kind = "individual"
    # FIXME we support group for the kind attribute but we do not support
    # member attributes in any way.
    _supported_kinds = ("individual", "group", "org", "location",
                        "application", "device")

    # vcard v3.0 supports the following type values
    phone_types_v3 = ("bbs", "car", "cell", "fax", "home", "isdn", "msg",
                      "modem", "pager", "pcs", "video", "voice", "work")
    email_types_v3 = ("home", "internet", "work", "x400")
    address_types_v3 = ("dom", "intl", "home", "parcel", "postal", "work")
    # vcard v4.0 supports the following type values
    phone_types_v4 = ("text", "voice", "fax", "cell", "video", "pager",
                      "textphone", "home", "work")
    email_types_v4 = ("home", "internet", "work")
    address_types_v4 = ("home", "work")

    def __init__(self, vcard: vobject.base.Component,
                 version: Optional[str] = None) -> None:
        """Initialize the wrapper around the given vcard.

        :param vcard: the vCard to wrap
        :param version: the version of the RFC to use (if the card has none)
        """
        self.vcard = vcard
        if not self.version:
            version = version or self._default_version
            logger.warning("Wrapping unversioned vCard object, setting "
                           "version to %s.", version)
            self.version = version
        elif self.version not in self._supported_versions:
            logger.warning("Wrapping vCard with unsupported version %s, this "
                           "might change any incompatible attributes.",
                           self.version)

    def __str__(self) -> str:
        return self.formatted_name

    @overload
    def get_first(self, property: Literal["n"]) -> Optional[vobject.vcard.Name]: ...
    @overload
    def get_first(self, property: Literal["adr"]) -> Optional[vobject.vcard.Address]: ...
    @overload
    def get_first(self, property: str) -> Optional[str]: ...
    def get_first(self, property: str) -> Union[None, str, vobject.vcard.Name,
                                                vobject.vcard.Address]:
        """Get a property from the underlying vCard.

        This method should only be called for properties with cardinality \\*1
        (zero or one).  Otherwise only the first value will be returned. If the
        property is not present None will be retuned.

        The type annotation for the return value is str but this is not
        enforced so it is up to the caller to make sure to only call this
        method for properties where the underlying vobject library returns a
        str.

        :param property: the field value to get
        :returns: the property value or None
        """
        try:
            return getattr(self.vcard, property).value
        except AttributeError:
            return None

    def get_all(self, name: str) -> list:
        """Get all values of the given vCard property.

        It does not matter what the individual vcard properties store as their
        value.  This function returns them untouched inside an aggregating
        list.

        If the property is part of a group containing exactly two items, with
        exactly one ABLABEL. the property will be prefixed with that ABLABEL.

        :param name: the name of the property (should be UPPER case)
        :returns: the values from all occurrences of the named property
        """
        try:
            values = getattr(self.vcard, f"{name.lower()}_list")
        except AttributeError:
            return []
        values = [{label: item.value} if (label := self._get_ablabel(item))
                  else item.value for item in values]
        return sorted(values, key=multi_property_key)

    def _delete_vcard_object(self, name: str) -> None:
        """Delete all fields with the given name from the underlying vCard.

        If a field that will be deleted is in a group with an X-ABLABEL field,
        that X-ABLABEL field will also be deleted.  These fields are commonly
        added by the Apple address book to attach custom labels to some fields.

        :param name: the name of the fields to delete
        """
        # first collect all vcard items, which should be removed
        to_be_removed = []
        for child in self.vcard.getChildren():
            if child.name == name:
                if child.group:
                    for label in self.vcard.getChildren():
                        if label.name == "X-ABLABEL" and \
                                label.group == child.group:
                            to_be_removed.append(label)
                to_be_removed.append(child)
        # then delete them one by one
        for item in to_be_removed:
            self.vcard.remove(item)

    @staticmethod
    def _parse_type_value(types: Sequence[str], supported_types: Sequence[str]
                          ) -> tuple[list[str], list[str], int]:
        """Parse type value of phone numbers, email and post addresses.

        :param types: list of type values
        :param supported_types: all allowed standard types
        :returns: tuple of standard and custom types and pref integer
        """
        custom_types = []
        standard_types = []
        pref = 0
        for type in types:
            type = type.strip()
            if type:
                if type.lower() in supported_types:
                    standard_types.append(type)
                elif type.lower() == "pref":
                    pref += 1
                elif re.match(r"^pref=\d{1,2}$", type.lower()):
                    pref += int(type.split("=")[1])
                else:
                    if type.lower().startswith("x-"):
                        custom_types.append(type[2:])
                        standard_types.append(type)
                    else:
                        custom_types.append(type)
                        standard_types.append("X-{}".format(type))
        return (standard_types, custom_types, pref)

    def _get_types_for_vcard_object(self, object: vobject.base.ContentLine,
                                    default_type: str) -> list[str]:
        """get list of types for phone number, email or post address

        :param object: vcard class object
        :param default_type: use if the object contains no type
        :returns: list of type labels
        """
        type_list = []
        # try to find label group for custom value type
        if object.group:
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group == object.group:
                    custom_type = label.value.strip()
                    if custom_type:
                        type_list.append(custom_type)
        # then load type from params dict
        standard_types = object.params.get("TYPE")
        if standard_types is not None:
            if not isinstance(standard_types, list):
                standard_types = [standard_types]
            for type in standard_types:
                type = type.strip()
                if type and type.lower() != "pref":
                    if not type.lower().startswith("x-"):
                        type_list.append(type)
                    elif type[2:].lower() not in [x.lower()
                                                  for x in type_list]:
                        # add x-custom type in case it's not already added by
                        # custom label for loop above but strip x- before
                        type_list.append(type[2:])
        # try to get pref parameter from vcard version 4.0
        try:
            type_list.append("pref={}".format(
                int(object.params.get("PREF")[0])))
        except (IndexError, TypeError, ValueError):
            # else try to determine, if type params contain pref attribute
            try:
                for type in object.params.get("TYPE"):
                    if type.lower() == "pref" and "pref" not in type_list:
                        type_list.append("pref")
            except TypeError:
                pass
        # return type_list or default type
        if type_list:
            return type_list
        return [default_type]

    @property
    def version(self) -> Optional[str]:
        return self.get_first("version")

    @version.setter
    def version(self, value: str) -> None:
        if value not in self._supported_versions:
            logger.warning("Setting vcard version to unsupported version %s",
                           value)
        # All vCards should only always have one version, this is a requirement
        # for version 4 but also makes sense for all other versions.
        self._delete_vcard_object("VERSION")
        version = self.vcard.add("version")
        version.value = convert_to_vcard("version", value, str)

    @property
    def uid(self) -> Optional[str]:
        return self.get_first("uid")

    @uid.setter
    def uid(self, value: str) -> None:
        # All vCards should only always have one UID, this is a requirement
        # for version 4 but also makes sense for all other versions.
        self._delete_vcard_object("UID")
        uid = self.vcard.add('uid')
        uid.value = convert_to_vcard("uid", value, str)

    def _update_revision(self) -> None:
        """Generate a new REV field for the vCard, replace any existing

        All vCards should only always have one revision, this is a
        requirement for version 4 but also makes sense for all other
        versions.
        """
        self._delete_vcard_object("REV")
        rev = self.vcard.add('rev')
        rev.value = datetime.datetime.now().strftime("%Y%m%dT%H%M%SZ")

    @property
    def birthday(self) -> Optional[Date]:
        """Return the birthday as a datetime object or a string depending on
        whether it is of type text or not.  If no birthday is present in the
        vcard None is returned.

        :returns: contacts birthday or None if not available
        """
        # vcard 4.0 could contain a single text value
        try:
            if self.vcard.bday.params.get("VALUE")[0] == "text":
                return self.vcard.bday.value
        except (AttributeError, IndexError, TypeError):
            pass
        # else try to convert to a datetime object
        try:
            return string_to_date(self.vcard.bday.value)
        except (AttributeError, ValueError):
            pass
        return None

    @birthday.setter
    def birthday(self, date: Date) -> None:
        """Store the given date as BDAY in the vcard.

        :param date: the new date to store as birthday
        """
        value, text = self._prepare_birthday_value(date)
        if value is None:
            logger.warning('Failed to set anniversary to %s', date)
            return
        bday = self.vcard.add('bday')
        bday.value = value
        if text:
            bday.params['VALUE'] = ['text']

    @property
    def anniversary(self) -> Optional[Date]:
        """
        :returns: contacts anniversary or None if not available
        """
        # vcard 4.0 could contain a single text value
        try:
            if self.vcard.anniversary.params.get("VALUE")[0] == "text":
                return self.vcard.anniversary.value
        except (AttributeError, IndexError, TypeError):
            pass
        # else try to convert to a datetime object
        try:
            return string_to_date(self.vcard.anniversary.value)
        except (AttributeError, ValueError):
            # vcard 3.0: x-anniversary (private object)
            try:
                return string_to_date(self.vcard.x_anniversary.value)
            except (AttributeError, ValueError):
                pass
        return None

    @anniversary.setter
    def anniversary(self, date: Date) -> None:
        value, text = self._prepare_birthday_value(date)
        if value is None:
            logger.warning('Failed to set anniversary to %s', date)
            return
        if text:
            anniversary = self.vcard.add('anniversary')
            anniversary.params['VALUE'] = ['text']
            anniversary.value = value
        elif self.version == "4.0":
            self.vcard.add('ANNIVERSARY').value = value
        else:
            self.vcard.add('X-ANNIVERSARY').value = value

    def _get_ablabel(self, item: vobject.base.ContentLine) -> str:
        """Get an ABLABEL for a specified item in the vCard.
        Will return the ABLABEL only if the item is part of a group with
        exactly two items, exactly one of which is an ABLABEL.

        :param item: the item to be labelled
        :returns: the ABLABEL in the circumstances above or an empty string
        """
        label = ""
        if item.group:
            count = 0
            for child in self.vcard.getChildren():
                if child.group and child.group == item.group:
                    count += 1
                    if child.name == "X-ABLABEL":
                        if label == "":
                            label = child.value
                        else:
                            return ""
            if count != 2:
                label = ""
        return label

    def _get_new_group(self, group_type: str = "") -> str:
        """Get an unused group name for adding new groups. Uses the form
        item123 or itemgroup_type123 if a grouptype is specified.

        :param group_type: (Optional) a string to add between "item" and
            the number
        :returns: the name of the first unused group of the specified form
        """
        counter = 1
        while True:
            group_name = "item{}{}".format(group_type, counter)
            for child in self.vcard.getChildren():
                if child.group and child.group == group_name:
                    counter += 1
                    break
            else:
                return group_name

    def _add_labelled_property(
            self, property: str, value: StrList, label: Optional[str] = None,
            name_groups: bool = False,
            allowed_object_type: Union[None, type[str], type[list]] = str) -> None:
        """Add an object to the VCARD. If a label is given it will be added to
        a group with an ABLABEL.

        :param property: type of object to add to the VCARD.
        :param value: Contents of the object to add.
        :param label: an optional label to add to the property
        :param name_groups: (Optional) If True, use the property in the
            group name for labelled objects.
        :param allowed_object_type: (Optional) set the accepted return type
            for vcard attribute
        """
        obj = self.vcard.add(property)
        obj.value = convert_to_vcard(property, value, allowed_object_type)
        if label:
            group_name = self._get_new_group(property if name_groups else "")
            obj.group = group_name
            ablabel_obj = self.vcard.add('X-ABLABEL')
            ablabel_obj.group = group_name
            ablabel_obj.value = label

    def _prepare_birthday_value(self, date: Date) -> tuple[Optional[str],
                                                           bool]:
        """Prepare a value to be stored in a BDAY or ANNIVERSARY attribute.

        :param date: the date like value to be stored
        :returns: the object to set as the .value for the attribute and whether
            it should be stored as plain text
        """
        if isinstance(date, str):
            if self.version == "4.0":
                return date.strip(), True
            return None, False
        tz = date.tzname()
        if date.year == 1900 and date.month != 0 and date.day != 0 \
                and date.hour == 0 and date.minute == 0 and date.second == 0 \
                and self.version == "4.0":
            fmt = '--%m%d'
        elif tz and tz[3:]:
            if self.version == "4.0":
                fmt = "%Y%m%dT%H%M%S{}".format(tz[3:])
            else:
                fmt = "%FT%T{}".format(tz[3:])
        elif date.hour != 0 or date.minute != 0 or date.second != 0:
            if self.version == "4.0":
                fmt = "%Y%m%dT%H%M%SZ"
            else:
                fmt = "%FT%TZ"
        else:
            if self.version == "4.0":
                fmt = "%Y%m%d"
            else:
                fmt = "%F"
        return date.strftime(fmt), False

    @property
    def kind(self) -> str:
        return self.get_first(self._kind_attribute_name().lower()) or self._default_kind

    @kind.setter
    def kind(self, value: str) -> None:
        if value == "":
            self._delete_vcard_object(self._kind_attribute_name().lower())
            return
        value = value.lower()
        if not any(k.startswith(value) for k in self._supported_kinds):
            raise ValueError(f"Bad value for kind attribute: {value}")
        self.vcard.add(self._kind_attribute_name()).value = value

    def _kind_attribute_name(self) -> str:
        return "{}KIND".format("" if self.version == "4.0" else "X-")

    @property
    def formatted_name(self) -> str:
        fn = self.get_first("fn")
        if fn:
            return fn
        self.formatted_name = ""
        return self.get_first("fn") or ""

    @formatted_name.setter
    def formatted_name(self, value: str) -> None:
        # TODO cardinality 1*
        """Set the FN field to the new value.

        All previously existing FN fields are deleted.  Version 4 of the specs
        requires the vCard to only have one FN field.  For other versions we
        enforce this equally.

        :param value: the new formatted name
        """
        self._delete_vcard_object("FN")
        if value:
            final = convert_to_vcard("FN", value, str)
        elif self._get_first_names() or self._get_last_names():
            # autofill the FN field from the N field
            names = [self._get_name_prefixes(), self._get_first_names(),
                     self._get_last_names(), self._get_name_suffixes()]
            names = [x for x in names if x]
            final = list_to_string(names, " ")
        else:  # add an empty FN
            final = ""
        self.vcard.add("FN").value = final

    def _get_names_part(self, part: str) -> list[str]:
        """Get some part of the "N" entry in the vCard as a list

        :param part: the name to get e.g. "prefix" or "given"
        :returns: a list of entries for this name part
        """
        try:
            the_list = getattr(self.vcard.n.value, part)
        except AttributeError:
            return []
        # check if list only contains empty strings
        if not ''.join(the_list):
            return []
        return the_list if isinstance(the_list, list) else [the_list]

    def _get_name_prefixes(self) -> list[str]:
        return self._get_names_part("prefix")

    def _get_first_names(self) -> list[str]:
        return self._get_names_part("given")

    def _get_additional_names(self) -> list[str]:
        return self._get_names_part("additional")

    def _get_last_names(self) -> list[str]:
        return self._get_names_part("family")

    def _get_name_suffixes(self) -> list[str]:
        return self._get_names_part("suffix")

    def get_first_name_last_name(self) -> str:
        """Compute the full name of the contact by joining first, additional
        and last names together
        """
        names = self._get_first_names() + self._get_additional_names() + \
            self._get_last_names()
        if names:
            return list_to_string(names, " ")
        return self.formatted_name

    def get_last_name_first_name(self) -> str:
        """Compute the full name of the contact by joining the last names and
        then after a comma the first and additional names together
        """
        last_names: list[str] = []
        if self._get_last_names():
            last_names += self._get_last_names()
        first_and_additional_names = self._get_first_names() + \
            self._get_additional_names()
        if last_names and first_and_additional_names:
            return "{}, {}".format(
                list_to_string(last_names, " "),
                list_to_string(first_and_additional_names, " "))
        if last_names:
            return list_to_string(last_names, " ")
        if first_and_additional_names:
            return list_to_string(first_and_additional_names, " ")
        return self.formatted_name

    @property
    def first_name(self) -> Optional[str]:
        if parts := self._get_first_names():
            return list_to_string(parts, " ")
        return None

    @property
    def last_name(self) -> Optional[str]:
        if parts := self._get_last_names():
            return list_to_string(parts, " ")
        return None

    def _add_name(self, prefix: StrList, first_name: StrList,
                  additional_name: StrList, last_name: StrList,
                  suffix: StrList) -> None:
        """Add an N entry to the vCard. No old entries are affected.

        :param prefix:
        :param first_name:
        :param additional_name:
        :param last_name:
        :param suffix:
        """
        name_obj = self.vcard.add('n')
        name_obj.value = vobject.vcard.Name(
            prefix=convert_to_vcard("name prefix", prefix, None),
            given=convert_to_vcard("first name", first_name, None),
            additional=convert_to_vcard("additional name", additional_name, None),
            family=convert_to_vcard("last name", last_name, None),
            suffix=convert_to_vcard("name suffix", suffix, None))

    @property
    def organisations(self) -> list[Union[list[str], dict[str, list[str]]]]:
        """
        :returns: list of organisations, sorted alphabetically
        """
        return self.get_all("org")

    def _add_organisation(self, organisation: StrList, label: Optional[str] = None) -> None:
        """Add one ORG entry to the underlying vcard

        :param organisation: the value to add
        :param label: an optional label to add
        """
        self._add_labelled_property("org", organisation, label, True, list)
        # check if fn attribute is already present
        if not self.vcard.getChildValue("fn") and self.organisations:
            # if not, set fn to organisation name
            first_org = self.organisations[0]
            if isinstance(first_org, dict):
                first_org = list(first_org.values())[0]
            org_value = list_to_string(first_org, ", ")
            self.formatted_name = org_value.replace("\n", " ").replace("\\",
                                                                       "")
            showas_obj = self.vcard.add('x-abshowas')
            showas_obj.value = "COMPANY"

    @property
    def titles(self) -> LabeledStrs:
        return self.get_all("title")

    def _add_title(self, title: str, label: Optional[str] = None) -> None:
        self._add_labelled_property("title", title, label, True)

    @property
    def roles(self) -> LabeledStrs:
        return self.get_all("role")

    def _add_role(self, role: str, label: Optional[str] = None) -> None:
        self._add_labelled_property("role", role, label, True)

    @property
    def nicknames(self) -> LabeledStrs:
        return self.get_all("nickname")

    def _add_nickname(self, nickname: str, label: Optional[str] = None) -> None:
        self._add_labelled_property("nickname", nickname, label, True)

    @property
    def notes(self) -> LabeledStrs:
        return self.get_all("note")

    def _add_note(self, note: str, label: Optional[str] = None) -> None:
        self._add_labelled_property("note", note, label, True)

    @property
    def webpages(self) -> LabeledStrs:
        return self.get_all("url")

    def _add_webpage(self, webpage: str, label: Optional[str] = None) -> None:
        self._add_labelled_property("url", webpage, label, True)

    @property
    def categories(self) -> Union[list[str], list[list[str]]]:
        category_list = self.get_all("categories")
        if not category_list:
            return category_list
        category_list = [value if isinstance(value, list) else [value] for
            value in category_list]
        if len(category_list) == 1:
            return category_list[0]
        return sorted(category_list)

    def _add_category(self, categories: list[str]) -> None:
        """Add categories to the vCard

        :param categories:
        """
        categories_obj = self.vcard.add('categories')
        categories_obj.value = convert_to_vcard("category", categories, list)

    @property
    def phone_numbers(self) -> dict[str, list[str]]:
        """
        :returns: dict of type and phone number list
        """
        phone_dict: dict[str, list[str]] = {}
        for child in self.vcard.getChildren():
            if child.name == "TEL":
                # phone types
                type = list_to_string(
                    self._get_types_for_vcard_object(child, "voice"), ", ")
                if type not in phone_dict:
                    phone_dict[type] = []
                # phone value
                #
                # vcard version 4.0 allows URI scheme "tel" in phone attribute value
                # Doc: https://tools.ietf.org/html/rfc6350#section-6.4.1
                # example: TEL;VALUE=uri;PREF=1;TYPE="voice,home":tel:+1-555-555-5555;ext=5555
                if child.value.lower().startswith("tel:"):
                    # cut off the "tel:" uri prefix
                    phone_dict[type].append(child.value[4:])
                else:
                    # free text field
                    phone_dict[type].append(child.value)
        # sort phone number lists
        for number_list in phone_dict.values():
            number_list.sort()
        return phone_dict

    def _add_phone_number(self, type: str, number: str) -> None:
        standard_types, custom_types, pref = self._parse_type_value(
            string_to_list(type, ","), self.phone_types_v4 if
            self.version == "4.0" else self.phone_types_v3)
        if not standard_types and not custom_types and pref == 0:
            raise ValueError(f"Label for phone number {number} is missing.")
        if len(custom_types) > 1:
            raise ValueError(f"Phone number {number} got more than one "
                             "custom label: " +
                             list_to_string(custom_types, ", "))
        phone_obj = self.vcard.add('tel')
        if self.version == "4.0":
            phone_obj.value = "tel:{}".format(
                convert_to_vcard("phone number", number, str))
            phone_obj.params['VALUE'] = ["uri"]
            if pref > 0:
                phone_obj.params['PREF'] = str(pref)
        else:
            phone_obj.value = convert_to_vcard("phone number", number, str)
            if pref > 0:
                standard_types.append("pref")
        if standard_types:
            phone_obj.params['TYPE'] = standard_types
        if custom_types:
            custom_label_count = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith(
                        "itemtel"):
                    custom_label_count += 1
            group_name = "itemtel{}".format(custom_label_count + 1)
            phone_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = custom_types[0]

    @property
    def emails(self) -> dict[str, list[str]]:
        """
        :returns: dict of type and email address list
        """
        email_dict: dict[str, list[str]] = {}
        try:
            emails = self.vcard.email_list
        except AttributeError:
            return {}
        for child in emails:
            type = list_to_string(
                self._get_types_for_vcard_object(child, "internet"), ", ")
            if type not in email_dict:
                email_dict[type] = []
            email_dict[type].append(child.value)
        # sort email address lists
        for email_list in email_dict.values():
            email_list.sort()
        return email_dict

    def add_email(self, type: str, address: str) -> None:
        standard_types, custom_types, pref = self._parse_type_value(
            string_to_list(type, ","), self.email_types_v4 if
            self.version == "4.0" else self.email_types_v3)
        if not standard_types and not custom_types and pref == 0:
            raise ValueError(f"Label for email address {address} is missing.")
        if len(custom_types) > 1:
            raise ValueError(f"Email address {address} got more "
                             "than one custom label: " +
                             list_to_string(custom_types, ", "))
        email_obj = self.vcard.add('email')
        email_obj.value = convert_to_vcard("email address", address, str)
        if self.version == "4.0":
            if pref > 0:
                email_obj.params['PREF'] = str(pref)
        else:
            if pref > 0:
                standard_types.append("pref")
        if standard_types:
            email_obj.params['TYPE'] = standard_types
        if custom_types:
            custom_label_count = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith(
                        "itememail"):
                    custom_label_count += 1
            group_name = "itememail{}".format(custom_label_count + 1)
            email_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = custom_types[0]

    @property
    def post_addresses(self) -> dict[str, list[PostAddress]]:
        """
        :returns: dict of type and post address list
        """
        post_adr_dict: dict[str, list[PostAddress]] = {}
        try:
            addresses = self.vcard.adr_list
        except AttributeError:
            return {}
        for child in addresses:
            type = list_to_string(self._get_types_for_vcard_object(
                child, "home"), ", ")
            if type not in post_adr_dict:
                post_adr_dict[type] = []
            post_adr_dict[type].append({"box": child.value.box,
                                        "extended": child.value.extended,
                                        "street": child.value.street,
                                        "code": child.value.code,
                                        "city": child.value.city,
                                        "region": child.value.region,
                                        "country": child.value.country})
        # sort post address lists
        for post_adr_list in post_adr_dict.values():
            post_adr_list.sort(key=lambda x: (
                list_to_string(x['city'], " ").lower(),
                list_to_string(x['street'], " ").lower()))
        return post_adr_dict

    def get_formatted_post_addresses(self) -> dict[str, list[str]]:
        formatted_post_adr_dict: dict[str, list[str]] = {}
        for type, post_adr_list in self.post_addresses.items():
            formatted_post_adr_dict[type] = []
            for post_adr in post_adr_list:
                def get(name: str) -> str:
                    return list_to_string(post_adr.get(name, ""), " ")

                # remove empty fields to avoid empty lines
                for x in list(post_adr.keys()):
                    if post_adr.get(x) == "":
                        del post_adr[x]

                strings = []
                if "street" in post_adr:
                    strings.append(list_to_string(
                        post_adr.get("street", ""), "\n"))
                if "box" in post_adr and "extended" in post_adr:
                    strings.append("{} {}".format(get("box"), get("extended")))
                elif "box" in post_adr:
                    strings.append(get("box"))
                elif "extended" in post_adr:
                    strings.append(get("extended"))
                if "code" in post_adr and "city" in post_adr:
                    strings.append("{} {}".format(get("code"), get("city")))
                elif "code" in post_adr:
                    strings.append(get("code"))
                elif "city" in post_adr:
                    strings.append(get("city"))
                if "region" in post_adr and "country" in post_adr:
                    strings.append("{}, {}".format(get("region"),
                                                   get("country")))
                elif "region" in post_adr:
                    strings.append(get("region"))
                elif "country" in post_adr:
                    strings.append(get("country"))
                formatted_post_adr_dict[type].append('\n'.join(strings))
        return formatted_post_adr_dict

    def _add_post_address(self, type: str, box: StrList, extended: StrList,
                          street: StrList, code: StrList, city: StrList,
                          region: StrList, country: StrList) -> None:
        standard_types, custom_types, pref = self._parse_type_value(
            string_to_list(type, ","), self.address_types_v4
            if self.version == "4.0" else self.address_types_v3)
        if not standard_types and not custom_types and pref == 0:
            raise ValueError(f"Label for post address {street} is missing.")
        if len(custom_types) > 1:
            raise ValueError(f"Post address {street} got more than one custom "
                             "label: " + list_to_string(custom_types, ", "))
        adr_obj = self.vcard.add('adr')
        adr_obj.value = vobject.vcard.Address(
            box=convert_to_vcard("box address field", box, None),
            extended=convert_to_vcard("extended address field", extended, None),
            street=convert_to_vcard("street", street, None),
            code=convert_to_vcard("post code", code, None),
            city=convert_to_vcard("city", city, None),
            region=convert_to_vcard("region", region, None),
            country=convert_to_vcard("country", country, None))
        if self.version == "4.0":
            if pref > 0:
                adr_obj.params['PREF'] = str(pref)
        else:
            if pref > 0:
                standard_types.append("pref")
        if standard_types:
            adr_obj.params['TYPE'] = standard_types
        if custom_types:
            custom_label_count = 0
            for label in self.vcard.getChildren():
                if label.name == "X-ABLABEL" and label.group.startswith(
                        "itemadr"):
                    custom_label_count += 1
            group_name = "itemadr{}".format(custom_label_count + 1)
            adr_obj.group = group_name
            label_obj = self.vcard.add('x-ablabel')
            label_obj.group = group_name
            label_obj.value = custom_types[0]


class YAMLEditable(VCardWrapper):
    """Conversion of vcards to YAML and updating the vcard from YAML"""

    def __init__(self, vcard: vobject.base.Component,
                 supported_private_objects: Optional[list[str]] = None,
                 version: Optional[str] = None, localize_dates: bool = False
                 ) -> None:
        """Initialize attributes needed for yaml conversions

        :param supported_private_objects: the list of private property names
            that will be loaded from the actual vcard and represented in this
            pobject
        :param version: the version of the RFC to use in this card
        :param localize_dates: should the formatted output of anniversary
            and birthday be localized or should the iso format be used instead
        """
        self.supported_private_objects = supported_private_objects or []
        self.localize_dates = localize_dates
        super().__init__(vcard, version)

    #####################
    # getters and setters
    #####################

    def _get_private_objects(self) -> dict[str, LabeledStrs]:
        supported = [x.lower() for x in self.supported_private_objects]
        private_objects: dict[str, LabeledStrs] = {}
        for child in self.vcard.getChildren():
            lower = child.name.lower()
            if lower.startswith("x-") and lower[2:] in supported:
                key_index = supported.index(lower[2:])
                key = self.supported_private_objects[key_index]
                if key not in private_objects:
                    private_objects[key] = []
                ablabel = self._get_ablabel(child)
                private_objects[key].append(
                    {ablabel: child.value} if ablabel else child.value)
        # sort private object lists
        for value in private_objects.values():
            value.sort(key=multi_property_key)
        return private_objects

    def _add_private_object(self, key: str, value: str) -> None:
        self._add_labelled_property('X-' + key.upper(), value)

    def get_formatted_anniversary(self) -> str:
        return self._format_date_object(self.anniversary, self.localize_dates)

    def get_formatted_birthday(self) -> str:
        return self._format_date_object(self.birthday, self.localize_dates)

    #######################
    # object helper methods
    #######################

    @staticmethod
    def _format_date_object(date: Optional[Date], localize: bool) -> str:
        if not date:
            return ""
        if isinstance(date, str):
            return date
        if date.year == 1900 and date.month != 0 and date.day != 0 \
                and date.hour == 0 and date.minute == 0 and date.second == 0:
            return date.strftime("--%m-%d")
        tz = date.tzname()
        if (tz and tz[3:]) or (date.hour != 0 or date.minute != 0
                               or date.second != 0):
            if localize:
                return date.strftime(locale.nl_langinfo(locale.D_T_FMT))
            utc_offset = -time.timezone / 60 / 60
            return date.strftime("%FT%T+{:02}:00".format(int(utc_offset)))
        if localize:
            return date.strftime(locale.nl_langinfo(locale.D_FMT))
        return date.strftime("%F")

    @staticmethod
    def _filter_invalid_tags(contents: str) -> str:
        for pat, repl in [('aim', 'AIM'), ('gadu', 'GADUGADU'),
                          ('groupwise', 'GROUPWISE'), ('icq', 'ICQ'),
                          ('xmpp', 'JABBER'), ('msn', 'MSN'),
                          ('yahoo', 'YAHOO'), ('skype', 'SKYPE'),
                          ('irc', 'IRC'), ('sip', 'SIP')]:
            contents = re.sub('X-messaging/'+pat+'-All', 'X-'+repl, contents,
                              flags=re.IGNORECASE)
        return contents

    @staticmethod
    def _parse_yaml(input: str) -> dict:
        """Parse a YAML document into a dictionary and validate the data to
        some degree.

        :param input: the YAML document to parse
        :returns: the parsed data structure
        """
        yaml_parser = YAML(typ='base')
        # parse user input string
        try:
            contact_data = yaml_parser.load(input)
        except (yaml.parser.ParserError, yaml.scanner.ScannerError,
                yaml.constructor.DuplicateKeyError) as err:
            raise ValueError(err)
        else:
            if not contact_data:
                raise ValueError("Found no contact information")

        # check for available data
        # at least enter name or organisation
        if not (contact_data.get("First name") or contact_data.get("Last name")
                or contact_data.get("Organisation")):
            raise ValueError("You must either enter a name or an organisation")
        return contact_data

    @staticmethod
    def _set_string_list(setter: Callable[[str, Optional[str]], None],
                         key: str, data: dict[str, Union[str, list[str]]]
                         ) -> None:
        """Pre-process a string or list and set each value with the given
        setter

        :param setter: the setter method to add a value to a card
        :param key:
        :param data:
        """
        value = data.get(key)
        if value:
            if isinstance(value, str):
                setter(value, None)
            elif isinstance(value, dict):
                label, val = list(value.items())[0]
                setter(val, label)
            elif isinstance(value, list):
                for val in value:
                    if val:
                        if isinstance(val, dict):
                            label, v = list(val.items())[0]
                            setter(v, label)
                        else:
                            setter(val, None)
            else:
                raise ValueError(
                    "{} must be a string or a list of strings".format(key))

    def _set_date(self, target: str, key: str, data: dict) -> None:
        new = data.get(key)
        if not new:
            return
        if not isinstance(new, str):
            raise ValueError(f"{key} must be a string object.")
        if re.match(r"^text[\s]*=.*$", new):
            if self.version == "4.0":
                v1 = ', '.join(x.strip() for x in re.split(r"text[\s]*=", new)
                               if x.strip())
                if v1:
                    setattr(self, target, v1)
                return
            raise ValueError(f"Free text format for {key.lower()} only usable "
                             "with vcard version 4.0.")
        if re.match(r"^--\d\d-?\d\d$", new) and self.version != "4.0":
            raise ValueError(
                f"{key} format --mm-dd and --mmdd only usable with "
                "vcard version 4.0. You may use 1900 as placeholder, if "
                "the year is unknown.")
        try:
            v2 = string_to_date(new)
            if v2:
                setattr(self, target, v2)
            return
        except ValueError:
            pass
        raise ValueError(f"Wrong {key.lower()} format or invalid date\n"
                         "Use format yyyy-mm-dd or "
                         "yyyy-mm-ddTHH:MM:SS")

    def update(self, input: str) -> None:
        """Update this vcard with some yaml input

        :param input: a yaml string to parse and then use to update self
        """
        contact_data = self._parse_yaml(input)
        # update rev
        self._update_revision()

        # name
        self._delete_vcard_object("N")
        # although the "n" attribute is not explicitly required by the vcard
        # specification,
        # the vobject library throws an exception, if it doesn't exist
        # so add the name regardless if it's empty or not
        self._add_name(
            contact_data.get("Prefix", ""), contact_data.get("First name", ""),
            contact_data.get("Additional", ""),
            contact_data.get("Last name", ""), contact_data.get("Suffix", ""))
        if "Formatted name" in contact_data:
            self.formatted_name = contact_data.get("Formatted name", "")
        if not self.formatted_name:
            # Trigger the auto filling code in the setter.
            self.formatted_name = ""

        # nickname
        self._delete_vcard_object("NICKNAME")
        self._set_string_list(self._add_nickname, "Nickname", contact_data)

        # organisation
        self._delete_vcard_object("ORG")
        self._delete_vcard_object("X-ABSHOWAS")
        self._set_string_list(self._add_organisation, "Organisation",
                              contact_data)

        # kind
        self._delete_vcard_object(self._kind_attribute_name())
        try:
            self.kind = contact_data["Kind"]
        except KeyError:
            pass

        # role
        self._delete_vcard_object("ROLE")
        self._set_string_list(self._add_role, "Role", contact_data)

        # title
        self._delete_vcard_object("TITLE")
        self._set_string_list(self._add_title, "Title", contact_data)

        # phone
        self._delete_vcard_object("TEL")
        phone_data = contact_data.get("Phone")
        if phone_data:
            if isinstance(phone_data, dict):
                for type, number_list in phone_data.items():
                    if isinstance(number_list, str):
                        number_list = [number_list]
                    if isinstance(number_list, list):
                        for number in number_list:
                            if number:
                                self._add_phone_number(type, number)
                    else:
                        raise ValueError(
                            "Got no number or list of numbers for the "
                            f"phone number type {type}")
            else:
                raise ValueError("Missing type value for phone number field")

        # email
        self._delete_vcard_object("EMAIL")
        email_data = contact_data.get("Email")
        if email_data:
            if isinstance(email_data, dict):
                for type, email_list in email_data.items():
                    if isinstance(email_list, str):
                        email_list = [email_list]
                    if isinstance(email_list, list):
                        for email in email_list:
                            if email:
                                self.add_email(type, email)
                    else:
                        raise ValueError(
                            "Got no email or list of emails for the "
                            f"email address type {type}")
            else:
                raise ValueError("Missing type value for email address field")

        # post addresses
        self._delete_vcard_object("ADR")
        address_data = contact_data.get("Address")
        if address_data:
            if isinstance(address_data, dict):
                for type, post_adr_list in address_data.items():
                    if isinstance(post_adr_list, dict):
                        post_adr_list = [post_adr_list]
                    if isinstance(post_adr_list, list):
                        for post_adr in post_adr_list:
                            if isinstance(post_adr, dict):
                                address_not_empty = False
                                for key, value in post_adr.items():
                                    if key in ["Box", "Extended", "Street",
                                               "Code", "City", "Region",
                                               "Country"] and value:
                                        address_not_empty = True
                                        break
                                if address_not_empty:
                                    self._add_post_address(
                                        type, post_adr.get("Box", ""),
                                        post_adr.get("Extended", ""),
                                        post_adr.get("Street", ""),
                                        post_adr.get("Code", ""),
                                        post_adr.get("City", ""),
                                        post_adr.get("Region", ""),
                                        post_adr.get("Country", ""))
                            else:
                                raise ValueError(
                                    f"One of the {type} type address "
                                    "list items does not contain an address")
                    else:
                        raise ValueError("Got no address or list of addresses "
                                         f"for the post address type {type}")
            else:
                raise ValueError("Missing type value for post address field")

        # categories
        self._delete_vcard_object("CATEGORIES")
        cat_data = contact_data.get("Categories")
        if cat_data:
            if isinstance(cat_data, str):
                self._add_category([cat_data])
            elif isinstance(cat_data, list):
                only_contains_strings = True
                for sub_category in cat_data:
                    if not isinstance(sub_category, str):
                        only_contains_strings = False
                        break
                # if the category list only contains strings, pack all of them
                # in a single CATEGORIES vcard tag
                if only_contains_strings:
                    self._add_category(cat_data)
                else:
                    for sub_category in cat_data:
                        if sub_category:
                            if isinstance(sub_category, str):
                                self._add_category([sub_category])
                            else:
                                self._add_category(sub_category)
            else:
                raise ValueError(
                    "Category must be a string or a list of strings")

        # urls
        self._delete_vcard_object("URL")
        self._set_string_list(self._add_webpage, "Webpage", contact_data)

        # anniversary
        self._delete_vcard_object("ANNIVERSARY")
        self._delete_vcard_object("X-ANNIVERSARY")
        self._set_date('anniversary', 'Anniversary', contact_data)

        # birthday
        self._delete_vcard_object("BDAY")
        self._set_date('birthday', 'Birthday', contact_data)

        # private objects
        for supported in self.supported_private_objects:
            self._delete_vcard_object("X-{}".format(supported.upper()))
        private_data = contact_data.get("Private")
        if private_data:
            if isinstance(private_data, dict):
                for key, value_list in private_data.items():
                    if key in self.supported_private_objects:
                        if isinstance(value_list, str):
                            value_list = [value_list]
                        if isinstance(value_list, list):
                            for value in value_list:
                                if value:
                                    self._add_private_object(key, value)
                        else:
                            raise ValueError("Got no value or list of values "
                                             f"for the private object {key}")
                    else:
                        raise ValueError(
                            f"Private object key {key} was changed.\n"
                            "Supported private keys: " + ', '.join(
                                self.supported_private_objects))
            else:
                raise ValueError("Private objects must consist of a "
                                 "key : value pair.")

        # notes
        self._delete_vcard_object("NOTE")
        self._set_string_list(self._add_note, "Note", contact_data)

    def to_yaml(self) -> str:
        """Convert this contact to a YAML string

        The conversion follows the implicit schema that is given by the
        internal YAML template of khard.

        :returns: a YAML representation of this contact
        """

        translation_table = {
            "Formatted name": self.formatted_name,
            "Kind": self.kind,
            "Prefix": self._get_name_prefixes(),
            "First name": self._get_first_names(),
            "Additional": self._get_additional_names(),
            "Last name": self._get_last_names(),
            "Suffix": self._get_name_suffixes(),
            "Nickname": self.nicknames,
            "Organisation": self.organisations,
            "Title": self.titles,
            "Role": self.roles,
            "Phone": helpers.yaml_dicts(
                self.phone_numbers, defaults=["cell", "home"]),
            "Email": helpers.yaml_dicts(
                self.emails, defaults=["home", "work"]),
            "Categories": self.categories,
            "Note": self.notes,
            "Webpage": self.webpages,
            "Anniversary":
                helpers.yaml_anniversary(self.anniversary, self.version or
                                         self._default_version),
            "Birthday":
                helpers.yaml_anniversary(self.birthday, self.version or
                                         self._default_version),
            "Address": helpers.yaml_addresses(
                self.post_addresses, ["Box", "Extended", "Street", "Code",
                    "City", "Region", "Country"], defaults=["home"])
        }
        template = helpers.get_new_contact_template()
        yaml = YAML()
        yaml.indent(mapping=4, sequence=4, offset=2)
        template_obj = yaml.load(template)
        for key in template_obj:
            value = translation_table.get(key, None)
            template_obj[key] = helpers.yaml_clean(value)

        if self.supported_private_objects:
            template_obj["Private"] = helpers.yaml_clean(
                helpers.yaml_dicts(
                    self._get_private_objects(),
                    self.supported_private_objects
                ))

        stream = io.StringIO()
        yaml.dump(template_obj, stream)
        # posix standard: eof char must be \n
        return stream.getvalue() + "\n"


@contextlib.contextmanager
def atomic_write(dest: str, overwrite: bool = False) -> Iterator[IO[str]]:
    """Atomically write to the destination file.

    Optionally overwrite the path (using rename) rather than using `os.link`.
    """
    fd, src = tempfile.mkstemp(prefix=os.path.basename(dest), dir=os.path.dirname(dest))
    file = os.fdopen(fd, mode='w')
    try:
        yield file
    except Exception:
        try:
            file.close()
            os.unlink(src)
        except Exception:
            pass
        raise
    else:
        file.flush()
        file.close()
        if overwrite:
            os.rename(src, dest)
        else:
            try:
                os.link(src, dest)
            finally:
                os.unlink(src)


class Contact(YAMLEditable):

    def __init__(self, vcard: vobject.base.Component,
                 address_book: "address_book.VdirAddressBook", filename: str,
                 supported_private_objects: Optional[list[str]] = None,
                 vcard_version: Optional[str] = None,
                 localize_dates: bool = False) -> None:
        """Initialize the vcard object.

        :param vcard: the vCard to wrap
        :param address_book: a reference to the address book where this vcard
            is stored
        :param filename: the path to the file where this vcard is stored
        :param supported_private_objects: the list of private property names
            that will be loaded from the actual vcard and represented in this
            object
        :param vcard_version: the version of the RFC to use
        :param localize_dates: should the formatted output of anniversary and
            birthday be localized or should the iso format be used instead
        """
        self.address_book = address_book
        self.filename = filename
        super().__init__(vcard, supported_private_objects, vcard_version,
                         localize_dates)

    #######################################
    # factory methods to create new contact
    #######################################

    @classmethod
    def new(cls, address_book: "address_book.VdirAddressBook",
            supported_private_objects: Optional[list[str]] = None,
            version: Optional[str] = None, localize_dates: bool = False
            ) -> "Contact":
        """Create a new Contact from scratch"""
        vcard = vobject.vCard()
        uid = helpers.get_random_uid()
        filename = os.path.join(address_book.path, uid + ".vcf")
        card = cls(vcard, address_book, filename, supported_private_objects,
                   version, localize_dates)
        card.uid = uid
        return card

    @classmethod
    def from_file(cls, address_book: "address_book.VdirAddressBook",
                  filename: str, query: Query = AnyQuery(),
                  supported_private_objects: Optional[list[str]] = None,
                  localize_dates: bool = False) -> Optional["Contact"]:
        """Load a Contact object from a .vcf file if the plain file
        matches the query.

        :param address_book: the address book where this contact is stored
        :param filename: the file name of the .vcf file
        :param query: the query to search in the source file or None to load
            the file unconditionally
        :param supported_private_objects: the list of private property names
            that will be loaded from the actual vcard and represented in this
            object
        :param localize_dates: should the formatted output of anniversary
            and birthday be localized or should the iso format be used instead
        :returns: the loaded Contact or None if the file didn't match
        """
        with open(filename, "r") as file:
            contents = file.read()
        if query.match(contents):
            try:
                vcard = vobject.readOne(contents)
            except Exception:
                logger.warning("Filtering some problematic tags from %s",
                               filename)
                # if creation fails, try to repair some vcard attributes
                vcard = vobject.readOne(cls._filter_invalid_tags(contents))
            return cls(vcard, address_book, filename,
                       supported_private_objects, None, localize_dates)
        return None

    @classmethod
    def from_yaml(cls, address_book: "address_book.VdirAddressBook", yaml: str,
                  supported_private_objects: Optional[list[str]] = None,
                  version: Optional[str] = None, localize_dates: bool = False
                  ) -> "Contact":
        """Use this if you want to create a new contact from user input."""
        contact = cls.new(address_book, supported_private_objects, version,
                          localize_dates=localize_dates)
        contact.update(yaml)
        return contact

    @classmethod
    def clone_with_yaml_update(cls, contact: "Contact", yaml: str,
                               localize_dates: bool = False
                               ) -> "Contact":
        """
        Use this if you want to clone an existing contact and replace its data
        with new user input in one step.
        """
        contact = cls(
            copy.deepcopy(contact.vcard), address_book=contact.address_book,
            filename=contact.filename,
            supported_private_objects=contact.supported_private_objects,
            localize_dates=localize_dates)
        contact.update(yaml)
        return contact

    ######################################
    # overwrite some default class methods
    ######################################

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Contact) and \
            self.pretty(False) == other.pretty(False)

    def __ne__(self, other: object) -> bool:
        return not self == other

    def pretty(self, verbose: bool = True) -> str:
        strings = []

        # Every vcard must have an FN field per the RFC.
        strings.append("Name: {}".format(self.formatted_name))
        # name
        if self._get_first_names() or self._get_last_names():
            names = self._get_name_prefixes() + self._get_first_names() + \
                self._get_additional_names() + self._get_last_names() + \
                self._get_name_suffixes()
            strings.append("Full name: {}".format(list_to_string(names, " ")))

        # kind
        if self.kind:
            strings.append("Kind: {}".format(self.kind))

        # organisation
        if self.organisations:
            strings += helpers.convert_to_yaml(
                "Organisation", self.organisations, 0, -1, False)

        # address book name
        if verbose:
            strings.append("Address book: {}".format(self.address_book))

        # kind
        if self.kind is not None:
            strings.append("Kind: {}".format(self.kind))

        # person related information
        if (self.birthday is not None or self.anniversary is not None
                or self.nicknames or self.roles or self.titles):
            strings.append("General:")
            if self.anniversary:
                strings.append("    Anniversary: {}".format(
                    self.get_formatted_anniversary()))
            if self.birthday:
                strings.append(
                    "    Birthday: {}".format(self.get_formatted_birthday()))
            if self.nicknames:
                strings += helpers.convert_to_yaml(
                    "Nickname", self.nicknames, 4, -1, False)
            if self.roles:
                strings += helpers.convert_to_yaml(
                    "Role", self.roles, 4, -1, False)
            if self.titles:
                strings += helpers.convert_to_yaml(
                    "Title", self.titles, 4, -1, False)

        # phone numbers
        if self.phone_numbers:
            strings.append("Phone")
            for type, number_list in sorted(self.phone_numbers.items(),
                                            key=lambda k: k[0].lower()):
                strings += helpers.convert_to_yaml(
                    type, number_list, 4, -1, False)

        # email addresses
        if self.emails:
            strings.append("E-Mail")
            for type, email_list in sorted(self.emails.items(),
                                           key=lambda k: k[0].lower()):
                strings += helpers.convert_to_yaml(
                    type, email_list, 4, -1, False)

        # post addresses
        if self.post_addresses:
            strings.append("Address")
            for type, post_adr_list in sorted(
                    self.get_formatted_post_addresses().items(),
                    key=lambda k: k[0].lower()):
                strings += helpers.convert_to_yaml(
                    type, post_adr_list, 4, -1, False)

        # private objects
        if self._get_private_objects().keys():
            strings.append("Private:")
            for object in self.supported_private_objects:
                if object in self._get_private_objects():
                    strings += helpers.convert_to_yaml(
                        object, self._get_private_objects().get(object), 4, -1,
                        False)

        # misc stuff
        if self.categories or self.webpages or self.notes or (verbose
                                                              and self.uid):
            strings.append("Miscellaneous")
            if verbose and self.uid:
                strings.append("    UID: {}".format(self.uid))
            if self.categories:
                strings += helpers.convert_to_yaml(
                    "Categories", self.categories, 4, -1, False)
            if self.webpages:
                strings += helpers.convert_to_yaml(
                    "Webpage", self.webpages, 4, -1, False)
            if self.notes:
                strings += helpers.convert_to_yaml(
                    "Note", self.notes, 4, -1, False)
        return '\n'.join(strings) + '\n'

    def write_to_file(self, overwrite: bool = False) -> None:
        # make sure, that every contact contains a uid
        if not self.uid:
            self.uid = helpers.get_random_uid()
        try:
            with atomic_write(self.filename, overwrite=overwrite) as f:
                f.write(self.vcard.serialize())
        except vobject.base.ValidateError as err:
            raise Cancelled(f"Vcard is not valid.\n{err}", 4)
        except OSError as err:
            raise Cancelled(f"Can't write\n{err}", 4)

    def delete_vcard_file(self) -> None:
        try:
            os.remove(self.filename)
        except OSError as err:
            logger.error("Can not remove vCard file: %s", err)

    @classmethod
    def get_properties(cls) -> list[str]:
        """Return the property names that are defined on this class."""
        return [name for name in dir(Contact)
                if isinstance(getattr(Contact, name), property)]



"""Command line parsing and configuration logic for khard"""

import argparse
import logging
import sys

from .actions import Actions
from .exceptions import ConfigError
from .contacts import Contact
from .config import Config
from .query import AndQuery, AnyQuery, parse
from .version import version as khard_version


logger = logging.getLogger(__name__)


class FieldsArgument:
    """A factory to create callable objects for add_argument's type= parameter.

    The object can parse comma separated strings into list of strings, and can
    also check if the single elements are spelled correctly.
    """

    def __init__(self, *choices: str, nested: bool = False) -> None:
        """Initialize the factory

        :param choices: the comma separated strings must be one of these
        :param nested: if this is true the comma separated strings may
            designate nested fields and only the first component (separated by
            a dot) must match on of the choices
        """
        self._choices = sorted(choices)
        self._nested = nested

    def __call__(self, argument: str) -> list[str]:
        ret = []
        for candidate in argument.split(","):
            candidate = candidate.lower()
            test = candidate.split('.')[0] if self._nested else candidate
            if test in self._choices:
                ret.append(candidate)
            else:
                choices = ', '.join('"{}"'.format(c) for c in self._choices)
                raise argparse.ArgumentTypeError(
                    '"{}" is not an accepted field. Accepted fields are {}.'
                    .format(test, choices))
        return ret


def create_parsers() -> tuple[argparse.ArgumentParser,
                              argparse.ArgumentParser]:
    """Create two argument parsers.

    The first parser is mainly used to find the config file which can than be
    used to set some default values on the second parser.  The second parser
    can parse the remainder of the command line with the subcommand and all
    further options and arguments.

    :returns: the two parsers for the first and the second parsing pass
    """
    # Create the base argument parser.  It will be reused for the first and
    # second round of argument parsing.
    base = argparse.ArgumentParser(
        description="Khard is a vcard address book for the console",
        formatter_class=argparse.RawTextHelpFormatter, add_help=False)
    base.add_argument("-c", "--config", help="config file to use")
    base.add_argument("--debug", action="store_true",
                      help="enable debug output")
    base.add_argument("--skip-unparsable", action="store_true",
                      help="skip unparsable vcard files")
    base.add_argument("-v", "--version", action="version",
                      version="Khard version {}".format(khard_version))

    # Create the first argument parser.  Its main job is to set the correct
    # config file.  The config file is needed to get the default command if no
    # subcommand is given on the command line.  This parser will ignore most
    # arguments, as they will be parsed by the second parser.
    first_parser = argparse.ArgumentParser(parents=[base])
    first_parser.add_argument('remainder', nargs=argparse.REMAINDER)

    # Create the main argument parser.  It will handle the complete command
    # line only ignoring the config and debug options as these have already
    # been set.
    parser = argparse.ArgumentParser(parents=[base])

    # create address book subparsers with different help texts
    default_addressbook_parser = argparse.ArgumentParser(add_help=False)
    default_addressbook_parser.add_argument(
        "-a", "--addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify one or several comma separated address book names to "
        "narrow the list of contacts")
    new_addressbook_parser = argparse.ArgumentParser(add_help=False)
    new_addressbook_parser.add_argument(
        "-a", "--addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify address book in which to create the new contact")
    copy_move_addressbook_parser = argparse.ArgumentParser(add_help=False)
    copy_move_addressbook_parser.add_argument(
        "-a", "--addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify one or several comma separated address book names to "
        "narrow the list of contacts")
    copy_move_addressbook_parser.add_argument(
        "-A", "--target-addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify target address book in which to copy / move the "
        "selected contact")
    merge_addressbook_parser = argparse.ArgumentParser(add_help=False)
    merge_addressbook_parser.add_argument(
        "-a", "--addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify one or several comma separated address book names to "
        "narrow the list of source contacts")
    merge_addressbook_parser.add_argument(
        "-A", "--target-addressbook", default=[],
        type=lambda x: [y.strip() for y in x.split(",")],
        help="Specify one or several comma separated address book names to "
        "narrow the list of target contacts")

    # create input file subparsers with different help texts
    email_header_input_file_parser = argparse.ArgumentParser(add_help=False)
    email_header_input_file_parser.add_argument(
        "-i", "--input-file", default="-",
        help="Specify input email header file name or use stdin by default")
    template_input_file_parser = argparse.ArgumentParser(add_help=False)
    template_input_file_parser.add_argument(
        "-i", "--input-file", default="-",
        help="Specify input template file name or use stdin by default")
    template_input_file_parser.add_argument(
        "--open-editor", "--edit", action="store_true", help="Open the "
        "default text editor after successful creation of new contact")

    # create sort subparser
    sort_parser = argparse.ArgumentParser(add_help=False)
    sort_parser.add_argument(
        "-d", "--display",
        choices=("first_name", "last_name", "formatted_name"),
        help="Display names in contact table by first or last name")
    sort_parser.add_argument(
        "-g", "--group-by-addressbook", action="store_true",
        help="Group contact table by address book")
    sort_parser.add_argument(
        "-r", "--reverse", action="store_true",
        help="Reverse order of contact table")
    sort_parser.add_argument(
        "-s", "--sort", choices=("first_name", "last_name", "formatted_name"),
        help="Sort contact table by first or last name")

    # create search subparsers
    default_search_parser = argparse.ArgumentParser(add_help=False)
    default_search_parser.add_argument(
        "-f", "--search-in-source-files", action="store_true",
        help="Look into source vcf files to speed up search queries in "
        "large address books. Beware that this option could lead "
        "to incomplete results.")
    default_search_parser.add_argument(
        "search_terms", nargs="*", metavar="search terms", type=parse,
        default=[], help="search in specified or all fields to find matching "
        "contact")
    merge_search_parser = argparse.ArgumentParser(add_help=False)
    merge_search_parser.add_argument(
        "-f", "--search-in-source-files", action="store_true",
        help="Look into source vcf files to speed up search queries in "
        "large address books. Beware that this option could lead "
        "to incomplete results.")
    merge_search_parser.add_argument(
        "-t", "--target-contact", "--target", type=parse,
        help="search for a matching target contact")
    merge_search_parser.add_argument(
        "source_search_terms", nargs="*", metavar="source", type=parse,
        default=[],
        help="search in specified or all fields to find matching source "
        "contact")

    # create subparsers for actions
    subparsers = parser.add_subparsers(dest="action", metavar="SUBCOMMAND")
    list_parser = subparsers.add_parser(
        "list",
        aliases=Actions.get_aliases("list"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="list all (selected) contacts",
        help="list all (selected) contacts")
    list_parser.add_argument(
        "-p", "--parsable", action="store_true",
        help="Machine readable format: uid\\tcontact_name\\taddress_book_name")
    field_argument = FieldsArgument(
        'index', 'name', 'phone', 'email', 'address_book',
        *Contact.get_properties(), nested=True)
    list_parser.add_argument(
        "-F", "--fields", default=[], type=field_argument,
        help="Comma separated list of fields to show "
        "(use -F help for a list of top level fields)")
    show_parser = subparsers.add_parser(
        "show",
        aliases=Actions.get_aliases("show"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="display detailed information about one contact",
        help="display detailed information about one contact")
    show_parser.add_argument(
        "--format", choices=("pretty", "yaml", "vcard"), default="pretty",
        help="select the output format")
    show_parser.add_argument(
        "-o", "--output-file", default=sys.stdout,
        type=argparse.FileType("w"),
        help="Specify output template file name or use stdout by default")
    subparsers.add_parser("template", help="print an empty yaml template")
    birthdays_parser = subparsers.add_parser(
        "birthdays",
        aliases=Actions.get_aliases("birthdays"),
        parents=[default_addressbook_parser, default_search_parser],
        description="list birthdays (sorted by month and day)",
        help="list birthdays (sorted by month and day)")
    birthdays_parser.add_argument(
        "-d", "--display",
        choices=("first_name", "last_name", "formatted_name"),
        help="Display names in birthdays table by first or last name")
    birthdays_parser.add_argument(
        "-p", "--parsable", action="store_true",
        help="Machine readable format: name\\tdate")
    email_parser = subparsers.add_parser(
        "email",
        aliases=Actions.get_aliases("email"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="list email addresses",
        help="list email addresses")
    email_parser.add_argument(
        "-p", "--parsable", action="store_true",
        help="Machine readable format: address\\tname\\ttype")
    email_parser.add_argument(
        "--remove-first-line", action="store_true",
        help="remove \"searching for '' ...\" line from parsable output "
        "(that line is required by mutt)")
    phone_parser = subparsers.add_parser(
        "phone",
        aliases=Actions.get_aliases("phone"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="list phone numbers",
        help="list phone numbers")
    phone_parser.add_argument(
        "-p", "--parsable", action="store_true",
        help="Machine readable format: number\\tname\\ttype")
    post_address_parser = subparsers.add_parser(
        "postaddress",
        aliases=Actions.get_aliases("postaddress"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="list postal addresses",
        help="list postal addresses")
    post_address_parser.add_argument(
        "-p", "--parsable", action="store_true",
        help="Machine readable format: address\\tname\\ttype")
    new_parser = subparsers.add_parser(
        "new",
        aliases=Actions.get_aliases("new"),
        parents=[new_addressbook_parser, template_input_file_parser],
        description="create a new contact",
        help="create a new contact")
    new_parser.add_argument(
        "--vcard-version", choices=("3.0", "4.0"), dest='preferred_version',
        help="Select preferred vcard version for new contact")
    add_email_parser = subparsers.add_parser(
        "add-email",
        aliases=Actions.get_aliases("add-email"),
        parents=[default_addressbook_parser, email_header_input_file_parser,
                 default_search_parser, sort_parser],
        description="Extract email address from the \"From:\" field of an "
        "email header and add to an existing contact or create a new one",
        help="Extract email address from the \"From:\" field of an email "
        "header and add to an existing contact or create a new one")
    add_email_parser.add_argument(
        "--vcard-version", choices=("3.0", "4.0"), dest='preferred_version',
        help="Select preferred vcard version for new contact")
    add_email_parser.add_argument(
        "-H", "--headers", default=["from"],
        type=lambda x: [y.lower() for y in x.split(",")],
        help="Extract contacts from the given comma separated header fields. "
        "`all` searches all headers.")
    add_email_parser.add_argument(
        "--skip-already-added", action="store_true",
        help="Skip already added email addresses")
    subparsers.add_parser(
        "merge",
        aliases=Actions.get_aliases("merge"),
        parents=[merge_addressbook_parser, merge_search_parser, sort_parser],
        description="merge two contacts",
        help="merge two contacts")
    edit_parser = subparsers.add_parser(
        "edit",
        aliases=Actions.get_aliases("edit"),
        parents=[default_addressbook_parser, template_input_file_parser,
                 default_search_parser, sort_parser],
        description="edit the data of a contact",
        help="edit the data of a contact")
    edit_parser.add_argument(
        "--format", choices=("yaml", "vcard"), default="yaml",
        help="specify the file format to use when editing")
    subparsers.add_parser(
        "copy",
        aliases=Actions.get_aliases("copy"),
        parents=[copy_move_addressbook_parser, default_search_parser,
                 sort_parser],
        description="copy a contact to a different addressbook",
        help="copy a contact to a different addressbook")
    subparsers.add_parser(
        "move",
        aliases=Actions.get_aliases("move"),
        parents=[copy_move_addressbook_parser, default_search_parser,
                 sort_parser],
        description="move a contact to a different addressbook",
        help="move a contact to a different addressbook")
    remove_parser = subparsers.add_parser(
        "remove",
        aliases=Actions.get_aliases("remove"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="remove a contact",
        help="remove a contact")
    remove_parser.add_argument(
        "--force", action="store_true",
        help="Remove contact without confirmation")
    subparsers.add_parser(
        "addressbooks",
        aliases=Actions.get_aliases("addressbooks"),
        description="list addressbooks",
        help="list addressbooks")
    subparsers.add_parser(
        "filename",
        aliases=Actions.get_aliases("filename"),
        parents=[default_addressbook_parser, default_search_parser,
                 sort_parser],
        description="list filenames of all matching contacts",
        help="list filenames of all matching contacts")

    # Replace the print_help method of the first parser with the print_help
    # method of the main parser.  This makes it possible to have the first
    # parser handle the help option so that command line help can be printed
    # without parsing the config file first (which is a problem if there are
    # errors in the config file).  The config file will still be parsed before
    # the full command line is parsed so errors in the config file might be
    # reported before command line syntax errors.
    first_parser.print_help = parser.print_help  # type: ignore

    return first_parser, parser


def parse_args(argv: list[str]) -> tuple[argparse.Namespace, Config]:
    """Parse the command line arguments and return the namespace that was
    creates by argparse.ArgumentParser.parse_args().

    :param argv: the command line arguments
    :returns: the namespace parsed from the command line
    """
    first_parser, parser = create_parsers()
    # Parse the command line with the first argument parser.  It will handle
    # the config option (its main job) and also the help, version and debug
    # options as these do not depend on anything else.
    args = first_parser.parse_args(argv)
    remainder = args.remainder

    # Set the log level to debug if given on the command line.  This is done
    # before parsing the config file to make it possible to debug the parsing
    # of the config file.
    if "debug" in args and args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Create the config instance.
    try:
        config = Config(args.config)
    except ConfigError as err:
        parser.exit(3, "Error in config file: {}\n".format(err))
    except OSError as err:
        parser.exit(3, "Error reading config file: {}\n".format(err))
    logger.debug("Finished parsing config=%s", vars(config))

    # Check the log level again and merge the value from the command line with
    # the config file.
    if ("debug" in args and args.debug) or config.debug:
        logging.basicConfig(level=logging.DEBUG)
    logger.debug("first args=%s", args)

    # Set the default command from the config file if none was given on the
    # command line.
    if not remainder or remainder[0] not in Actions.get_all():
        if config.default_action is None:
            parser.error("Missing subcommand on command line or default action"
                         " parameter in config.")
        remainder.insert(0, config.default_action)
        logger.debug("updated remainder=%s", remainder)

    # Save the last option that needs to be carried from the first parser run
    # to the second.
    skip = args.skip_unparsable

    # Parse the remainder of the command line.  All options from the previous
    # run have already been processed and are not needed any more.
    args = parser.parse_args(remainder)

    # Restore settings that are left from the first parser run.
    args.skip_unparsable = skip
    logger.debug("second args=%s", args)

    # Build conjunctive queries.
    if "source_search_terms" in args:
        args.source_search_terms = AndQuery.reduce(args.source_search_terms)
    if "search_terms" in args:
        args.search_terms = AndQuery.reduce(args.search_terms)
    if "target_contact" in args:
        # Only one of target_contact or target_uid can be set.
        args.target_contact = args.target_contact or AnyQuery()

    return args, config


def merge_args_into_config(args: argparse.Namespace, config: Config) -> Config:
    """Merge the parsed arguments from argparse into the config object.

    :param args: the parsed command line arguments
    :param config: the parsed config file
    :returns: the merged config object
    """
    config.merge_args(args)
    # Now we can safely initialize the address books as all command line
    # options have been incorporated into the config object.
    config.init_address_books()
    # If the user could but did not specify address books on the command line
    # it means they want to use all address books in that place.
    if "addressbook" in args and not args.addressbook:
        args.addressbook = [abook.name for abook in config.abooks]
    if "target_addressbook" in args and not args.target_addressbook:
        args.target_addressbook = [abook.name for abook in config.abooks]
    return config


def init(argv: list[str]) -> tuple[argparse.Namespace, Config]:
    """Initialize khard by parsing the command line and reading the config file

    :param argv: the command line arguments
    :returns: the parsed command line and the fully initialized config
    """
    args, conf = parse_args(argv)

    # if args.action isn't one of the defined actions, it must be an alias
    if args.action not in Actions.get_actions():
        # convert alias to corresponding action
        # example: "ls" --> "list"
        args.action = Actions.get_action(args.action)

    try:
        return args, merge_args_into_config(args, conf)
    except ConfigError as err:
        sys.exit(str(err))


def _sphinxarg_helper():
    """Helper function to return the correct argument parser for the docs."""
    return create_parsers()[1]
